import os
import yaml
import boto3
import requests
import tempfile
from enum import Enum
from pathlib import Path
from fmbench import defaults
from datetime import datetime
from typing import Optional, Dict
import importlib.resources as pkg_resources


FMBENCH_PACKAGE_NAME: str = "fmbench"

# This is the hf prefix to the source data file which acts as an identifier to whether
# the dataset is a hugging face dataset or not
HF_DATASET_PREFIX: str = "hf:"
DEFAULT_IMAGE_FORMAT: str = "JPEG"

# This is the ds_N default value of the number of rows to be processed from the hf dataset.
# If this value is already given in the 'dataset' section of the config file, that will be used.
DEFAULT_HF_DS_N_VALUE: int = 100

current_working_directory: str = Path.cwd()

CONFIG_FILEPATH_FILE: str = current_working_directory / 'config_filepath.txt'

PRICING_FALLBACK_YAML_PATH="https://raw.githubusercontent.com/aws-samples/foundation-model-benchmarking-tool/refs/heads/main/src/fmbench/configs/pricing_fallback.yml"

# S3 client initialization
s3_client = boto3.client('s3')
session = boto3.session.Session()
region_name = session.region_name
if region_name is None:
    print(f"boto3.session.Session().region_name is {region_name}, "
          f"going to use an metadata api to determine region name")
    # THIS CODE ASSUMED WE ARE RUNNING ON EC2, for everything else
    # the boto3 session should be sufficient to retrieve region name
    resp = requests.put("http://169.254.169.254/latest/api/token",
                        headers={"X-aws-ec2-metadata-token-ttl-seconds": "21600"})
    token = resp.text
    region_name = requests.get("http://169.254.169.254/latest/meta-data/placement/region",
                               headers={"X-aws-ec2-metadata-token": token}).text
    print(f"region_name={region_name}, also setting the AWS_DEFAULT_REGION env var")
    os.environ["AWS_DEFAULT_REGION"] = region_name
print(f"region_name={region_name}")

# Configuring the role ARN -- extract the role name
caller = boto3.client('sts').get_caller_identity()
account_id = caller.get('Account')
role_arn_from_env = os.environ.get('FMBENCH_ROLE_ARN')
if role_arn_from_env:
    print(f"role_arn_from_env={role_arn_from_env}, using it to set arn_string")
    arn_string = role_arn_from_env
else:
    print(f"role_arn_from_env={role_arn_from_env}, using current sts caller identity to set arn_string")
    arn_string = caller.get('Arn')
    # if this is an assumed role then remove the assumed role related pieces
    # because we are also using this role for deploying the SageMaker endpoint
    # arn:aws:sts::015469603702:assumed-role/SSMDefaultRoleForOneClickPvreReporting/i-0c5bba16a8b3dac51
    # should be converted to arn:aws:iam::015469603702:role/SSMDefaultRoleForOneClickPvreReporting
    if ":assumed-role/" in arn_string:
        role_name = arn_string.split("/")[-2]
        arn_string = f"arn:aws:iam::{account_id}:role/{role_name}"
        print(f"the sts role is an assumed role, setting arn_string to {arn_string}")
    else:
        arn_string = caller.get('Arn')

ROLE_NAME = arn_string.split('/')[-1]

current_config_file = os.environ.get("CONFIG_FILE_FMBENCH")
# if var is true, use that from cli
if current_config_file is not None:
    CONFIG_FILE = current_config_file
    print(f"config file current -> {CONFIG_FILE}, {current_config_file}")
else:
    CONFIG_FILE = Path(CONFIG_FILEPATH_FILE).read_text().strip()
    print(f"config file current -> {CONFIG_FILE}, {current_config_file}")


if CONFIG_FILE.startswith("s3://"):
    # Parse the S3 URL
    bucket_name, key_path = CONFIG_FILE.replace("s3://", "").split("/", 1)
    # Use boto3 to access the S3 service
    s3 = boto3.resource('s3')
    # Fetch the object from S3
    obj = s3.Object(bucket_name, key_path)
    # Read the object's content
    CONFIG_FILE_CONTENT = obj.get()['Body'].read().decode('utf-8')
elif CONFIG_FILE.startswith("https://"):
    # Fetch the content from the HTTPS URL
    response = requests.get(CONFIG_FILE)
    response.raise_for_status()  # Ensure we got a successful response
    CONFIG_FILE_CONTENT = response.text
else:
    CONFIG_FILE_CONTENT = Path(CONFIG_FILE).read_text()

# check if the file is still parameterized and if so replace the parameters with actual values
# if the file is not parameterized then the following statements change nothing
write_bucket = os.environ.get("WRITE_BUCKET", f"{defaults.DEFAULT_BUCKET_WRITE}-{region_name}-{account_id}")
# check if the tmp dir is used as an argument if local mode is set to yes. If so, then use that as the temp file directory
# else use the default `tempfile` option
tmp_dir = os.environ.get("TMP_DIR", tempfile.gettempdir())
args = dict(region=session.region_name,
            role_arn=arn_string,
            read_tmpdir=os.path.join(tmp_dir, defaults.DEFAULT_LOCAL_READ),
            write_tmpdir=os.path.join(tmp_dir, defaults.DEFAULT_LOCAL_WRITE),
            write_bucket=write_bucket,
            read_bucket=f"{defaults.DEFAULT_BUCKET_READ}-{region_name}-{account_id}")
CONFIG_FILE_CONTENT = CONFIG_FILE_CONTENT.format(**args)

# Load the configuration
config = yaml.safe_load(CONFIG_FILE_CONTENT)
local_mode = os.environ.get("LOCAL_MODE")
if local_mode == "yes":
    print("globals.py, local_mode = yes")
    config['aws']['s3_and_or_local_file_system'] = 'local'
    config['s3_read_data']['s3_or_local_file_system'] = 'local'
    if config['s3_read_data'].get('local_file_system_path') is None:
        config['s3_read_data']['local_file_system_path'] = os.path.join(tmp_dir, defaults.DEFAULT_LOCAL_READ)
    if config['aws'].get('local_file_system_path') is None:
        config['aws']['local_file_system_path'] = os.path.join(tmp_dir, defaults.DEFAULT_LOCAL_WRITE)

# iterate through each experiment and populate the parameters section in the inference spec
for i in range(len(config['experiments'])):
    # for the experiment at index i, look up the parameter set
    # retrieve the parameter set from the inference_parameter section
    # assign the parameters from that parameter set to a new key called
    # parameters in that experiment
    parameters = config['inference_parameters'][config['experiments'][i]['inference_spec']['parameter_set']]
    config['experiments'][i]['inference_spec']['parameters'] = parameters
    if config['experiments'][i].get('bucket') is None:
        config['experiments'][i]['bucket'] = config['aws']['bucket']
print(f"loaded config: {config}")


# get the model evaluation configuration file which contains information on the 
# ground truth, the method name, and directory structure being used
eval_config: Optional[Dict] = None
config_dir = Path(pkg_resources.files('fmbench'), 'configs')
# load the model evaluation configuration file based on the ground truth, 
# formatted into it from the main config file if any
if 'model_evaluations' in config and config['model_evaluations'] is not None:
    model_evaluation_common_file: str = config['model_evaluations']
    ground_truth_col_key: Optional[str] = config['datasets'].get('ground_truth_col_key', None)
    eval_module = Path(model_evaluation_common_file)
    eval_file_path: str = os.path.join(config_dir, eval_module)
    if ground_truth_col_key is not None:
        with open(eval_file_path, 'r') as file:
            model_eval_info = file.read()
            # load the preliminary unformatted config file to fetch the method name and plug it into
            # the prompt template file names
            eval_config = yaml.safe_load(model_eval_info)
            print(f"loaded eval configuration file: {eval_config}")
    else:
        eval_config=None
        print(f"Evalaution configuration file not found in the config file. Provide a valid eval configuration file name.")

# data directory and prompts
PER_ACCOUNT_DIR: str = f"{config['general']['name']}-{ROLE_NAME}"
DATA_DIR: str = os.path.join(PER_ACCOUNT_DIR, config['dir_paths']['data_prefix'])
PROMPTS_DIR = os.path.join(DATA_DIR, config['dir_paths']['prompts_prefix'])

# Metrics directory based on date and time
current_time = datetime.now()

# Assuming current_time is a datetime object
formatted_time = current_time.strftime("%Y/%m/%d/%H/%M")

# Split the formatted_time into components
year, month, day, hour, minute = formatted_time.split('/')

# Construct the METRICS_DIR path
METRICS_DIR = f"{DATA_DIR}/metrics/yyyy={year}/mm={month}/dd={day}/hh={hour}/mm={minute}"

METRICS_PER_INFERENCE_DIR = os.path.join(METRICS_DIR, "per_inference")
METRICS_PER_CHUNK_DIR = os.path.join(METRICS_DIR, "per_chunk")
METRICS_PER_POLL_EVAL_DIR_NAME: str = "per_poll_eval"

METRICS_PER_INFERENCE_DIR = os.path.join(METRICS_DIR, "per_inference")
METRICS_PER_CHUNK_DIR = os.path.join(METRICS_DIR, "per_chunk")
ENDPOINT_METRICS_FNAME = "endpoint_metrics.csv"
ENDPOINT_METRICS_SUMMARIZED_FNAME = "endpoint_metrics_summarized.csv"

# These are the column names that are present in the SageMaker and EC2 instance utilization metrics
# SageMaker utilization metrics contain an EndpointName column
SAGEMAKER_EP_NAME_COL: str = "EndpointName"
# These are the list of quantiles that are measured as a part of the metric utilization calculation
# for EC2 benchmarking
UTILIZATION_QUANTILE_METRICS = [0, 0.25, 0.5, 0.75, 1.0]

# Models directory based on date and time 
MODELS_DIR = f"{DATA_DIR}/models"

# Use this to upload to the s3 bucket (extracted from the config file)
BUCKET_NAME = config['aws']['bucket']
READ_BUCKET_NAME = config['s3_read_data']['read_bucket']

# S3 prefix
PREFIX_NAME = config['dir_paths']['data_prefix']

# SOURCE data is where your actual data resides in s3
SOURCE_DATA = config['s3_read_data']['source_data_prefix']

# Read the prompt template that the user uploads
PROMPT_TEMPLATE_S3_PREFIX = config['s3_read_data']['prompt_template_dir']

# Initialize the scripts directory
SCRIPTS_DIR: str = "fmbench/scripts"

# Contruct the path to the evaluation prompt and the different rules in 
# the rules directory for respective subjective eval criteria
if eval_config is not None:
    EVAL_PROMPT_TEMPLATES: str = os.path.join(PROMPT_TEMPLATE_S3_PREFIX,
                                              eval_config['model_evaluations']['model_eval_dir'].get('eval_prompts_dir', None))
    EVAL_DIR: str = eval_config['model_evaluations']['model_eval_dir'].get('eval_prompts_dir', None)
    EVAL_INSTRUCTIONS_DIR: str = eval_config['model_evaluations']['model_eval_dir'].get('eval_instructions_dir', None)

# METADATA DIR TO HANDLE DYNAMIC S3 PATHS FOR METRICS/RESULTS
METADATA_DIR:str = config['dir_paths']['metadata_dir']
METRICS_PATH_FNAME: str = "metrics_path.txt"

# Name of the .txt file where the HF token is stored
HF_TOKEN_FNAME: str = "hf_token.txt"

DIR_LIST = [DATA_DIR, PROMPTS_DIR, METRICS_DIR, MODELS_DIR, METRICS_PER_INFERENCE_DIR, METRICS_PER_CHUNK_DIR]

# this is for custom tokenizers
TOKENIZER_DIR_S3 = config['s3_read_data']['tokenizer_prefix']
TOKENIZER = 'tokenizer'
# we take the first experiment's model id as the model whose tokenizer is used for 
# counting tokens on the dataset. This is done just for ease of coding, this is just
# token counting logic on the client side (does not impact the tokenizer the model uses)
# NOTE: if tokenizer files are provided in the tokenizer directory then they take precedence
# if the files are not present then we load the tokenizer for this model id from Hugging Face
TOKENIZER_MODEL_ID = config['experiments'][0].get('model_id')

# If the model that is being benchmarked does not necessarilly have an HF model id, but you
# want to use the hf tokenizer for that model, then mention the "hf_tokenizer_model_id" in the
# experiments section to load the tokenizer at runtime. The alternative option to this is
# to provide the "config.json" and "tokenizer.json" files in the read directory. If neither are provided, 
# then the default 750-1000 tokens tokenizer will be used.
HF_TOKENIZER_MODEL_ID = config['experiments'][0].get('hf_tokenizer_model_id')

DEPLOYMENT_SCRIPT_S3 = config['s3_read_data']['scripts_prefix']

_ = list(map(lambda x: os.makedirs(x, exist_ok=True), DIR_LIST))

# Define the endpoint list as the config-general name plus the role arn for unique generation 
# from different roles in the same/different accounts
ENDPOINT_LIST_PATH: str = os.path.join(MODELS_DIR, "endpoints.json")

REQUEST_PAYLOAD_FPATH: str = os.path.join(PROMPTS_DIR, "payload.jsonl")
RESULTS_FPATH: str = os.path.join(METRICS_DIR, "results.csv")


class TRUNCATE_POLICY(str, Enum):
    AT_PROMPT_TOKEN_LENGTH = 'at-prompt-token-length'

# misc. metrics related
PLACE_HOLDER: int = -1705338041
RESULTS_DIR: str = f"results-{config['general']['name']}"

# benchmarking - metric filenames
COUNTS_FNAME: str = "experiment_counts.csv"
ERROR_RATES_FNAME: str = "error_rates.csv"
RESULTS_DESC_MD_FNAME: str = "report.md"
SUMMARY_METRICS_W_PRICING_FNAME: str = "summary_metrics_w_pricing.csv"
INSTANCE_PRICING_PER_HOUR_FNAME: str = "instance_pricing_per_hour.csv"
SUMMARY_METRICS_FOR_DATASET_W_SCORES_FNAME: str = "summary_metrics_for_dataset_w_scores.csv"
SUMMARY_METRICS_FOR_DATASET_W_SCORES_BEST_OPTION_FNAME: str = "summary_metrics_for_dataset_best_option.csv"
SUMMARY_METRICS_FOR_DATASET_W_SCORES_BEST_OPTION_EACH_INSTANCE_TYPE_FNAME: str = "summary_metrics_for_dataset_best_option_each_instance_type.csv"
SUMMARY_MODEL_ENDPOINT_COST_PER_INSTANCE: str = "endpoint_per_instance_per_run_costs.csv"
BUSINESS_SUMMARY_PLOT_FNAME: str = "business_summary.png"
BUSINESS_SUMMARY_PLOT_FNAME2: str = "business_summary_barchart.png"
LATENCY_CHART_PLOT_FNAME: str = "latency_summary_chart.png"

# evaluation - metric filenames
PER_INFERENCE_FILE_WITH_COSINE_SIMILARITY_SCORES: str = "per_inference_quantitative_eval_metrics.csv"
EVAL_COL_SUFFIX: str = '_eval_prompt'
PROCESSED_EVAL_PROMPT_PAYLOADS: str = "processed_eval_prompts_for_inference.csv"
MODEL_EVALUATION_JUDGE_COMPLETIONS_DIR: str = "judge_model_eval_completions"
MODEL_EVAL_COMPLETIONS_CSV: str = "raw_llm_as_a_judge_evals.csv"
LLM_JUDGE_PANEL_RESPONSE_SUMMARIES: str = "llm_as_a_judge_per_eval_summary.csv"
# this csv contains all of the incorrect verdict responses from the PoLL
# evaluation of responses using Max Voting. View this csv to get more insight
# into where the model went wrong, and what to fix
VERDICT_TYPE_BREAKDOWN_FOR_CORRECT_FILE: str = "verdict_type_breakdown_for_correct.csv"
VERDICT_TYPE_BREAKDOWN_FOR_INCORRECT_FILE: str = "verdict_type_breakdown_for_incorrect.csv"
PER_MODEL_ACCURACY_W_VERDICT_TYPE_FILE: str = "per_model_accuracy_w_verdict_type.csv"
MAJORITY_VOTE_DF_RAW_RESULTS_FILE: str = "majority_vote_results_raw.csv"
PER_PAYLOAD_MODEL_ACCURACY_MAJORITY_VOTING: str = "per_payload_model_accuracy_majority_vote.csv"
PER_MODEL_ACCURACY_PER_EVAL_JUDGE: str = "per_model_per_eval_judge_accuracy.csv"
CANDIDATE_MODEL_ACCURACY_FILE: str = "candidate_model_accuracy.csv"
INCORRECT_VERDICT_RESPONSES_FILE: str = "incorrect_verdict_responses.csv"
CORRECT_VERDICT_RESPONSES_FILE: str = "correct_verdict_responses.csv"
SCORING_RESULT_COUNT_POLL: str = "PoLL_result_count_correct_incorrect.csv"
PER_MODEL_ACCURACY_POLL: str = "PoLL_per_model_accuracy.csv"
PER_PAYLOAD_PER_MODEL_POLL_ACCURACY: str = "majority_vote_accuracy_per_payload_file.csv"
EVAL_COST_PER_JUDGE_MODEL: str = "eval_cost_per_llm_evaluator.csv"
# contains all tt data of the LLM completion from the evaluation process
ALL_EVALUATIONS_IN_TXT: str = "all_judges_evals.txt"
# contains the final analysis done by a final LLM in the loop to summarize
# all evaluations done by panel of LLM evaluators on candidate model responses
NEEDS_FURTHER_EVAL_FILE: str = "responses_need_further_eval.txt"
# accuracy charts 
PER_PAYLOAD_FILE_ACCURACY_TRAJECTORY: str = "accuracy_trajectory_per_payload.png"
OVERALL_CANDIDATE_MODEL_MAJORITY_VOTING_ACCURACY: str = "overall_candidate_model_majority_voting_accuracy.png"

# plot filenames
ERROR_RATES_PLOT_TEXT: str = "Error rates for different concurrency levels and instance types"
ERROR_RATES_PLOT_FNAME: str = "error_rates.png"
TOKENS_VS_LATENCY_PLOT_TEXT: str = "Tokens vs latency for different concurrency levels and instance types"
TOKENS_VS_LATENCY_PLOT_FNAME: str = "tokens_vs_latency.png"
CONCURRENCY_VS_INFERENCE_LATENCY_PLOT_FNAME: str = "concurrency_vs_inference_latency.png"
CONCURRENCY_VS_INFERENCE_LATENCY_PLOT_TEXT: str = "Concurrency Vs latency for different instance type for selected dataset"
LATENCY_BUDGET: int = 5

OVERALL_RESULTS_MD: str = """
# {title}

|**Last modified (UTC)** | **FMBench version**  |
|---|---|
|{dttm}|{fmbench_version}|


## Summary

{business_summary}

## Per instance results

The following table provides the best combinations for running inference for different sizes prompts on different instance types. The following dataset(s) were used for this test: {datasets}.

|Dataset   | Instance type   | Recommendation   |
|---|---|---|
"""

# Dataset=`{dataset}`, instance_type=`{instance_type}`
RESULT_DESC: str = """The best option for staying within a latency budget of `{latency_budget} seconds` on a `{instance_type}` for the `{dataset}` dataset is a `concurrency level of {concurrency}`. A concurrency level of {concurrency} achieves an `median latency of {latency_median} seconds`, for an `average prompt size of {prompt_size} tokens` and `completion size of {completion_size} tokens` with `{tpm} transactions/minute`."""

RESULT_ROW: str = "|`{dataset}`|`{instance_type}`|{desc}|"

RESULT_FAILURE_DESC: str = """This experiment did not find any combination of concurrency level and other configuration settings that could provide a response within a latency budget of `{latency_budget} seconds` on a `{instance_type}` for the `{dataset}` dataset."""

PROMPT_TEMPLATE: str = """<s>[INST] <<SYS>>
You are an assistant for question-answering tasks. Use the following pieces of retrieved context in the section demarcated by "```" to answer the question. If you don't know the answer just say that you don't know. Use three sentences maximum and keep the answer concise.
<</SYS>>

```
{context}
```

Question: {input}

[/INST]
Answer:

"""



# Import necessary libraries
import os
import json
import time
import boto3
import fmbench
import logging
import sagemaker
from typing import Dict
from pathlib import Path
from fmbench.scripts import constants
from sagemaker.huggingface import HuggingFaceModel
from sagemaker.huggingface import get_huggingface_llm_image_uri

# globals
HF_TOKEN_FNAME: str = os.path.join(os.path.dirname(os.path.realpath(__file__)), "hf_token.txt")

# Initialize the platform where this script deploys the model
PLATFORM: str = constants.PLATFORM_SAGEMAKER

# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Initialize your S3 client for your model uploading
s3_client = boto3.client('s3')

# Initialize the sagemaker and sagemaker runtime clients 
sm_client = boto3.client("sagemaker")
smr_client = boto3.client("sagemaker-runtime")

tag = [
    {
        'Key': 'fmbench-version',
        'Value': fmbench.__version__
    }
]

# Function to create the llm hugging face model
def create_hugging_face_model(experiment_config: Dict, role_arn: str) -> HuggingFaceModel:
    # Define Model and Endpoint configuration parameter
    model_config = {
    'HF_MODEL_ID': experiment_config['model_id'],
    'SM_NUM_GPUS': json.dumps(experiment_config['env']['NUMBER_OF_GPU']), # Number of GPU used per replica
    'MAX_INPUT_LENGTH': json.dumps(4090),  # Max length of input text
    'MAX_TOTAL_TOKENS': json.dumps(4096),  # Max length of the generation (including input text)
    'MAX_BATCH_TOTAL_TOKENS': json.dumps(8192),  # Limits the number of tokens that can be processed in parallel during the generation
    'HUGGING_FACE_HUB_TOKEN': Path(HF_TOKEN_FNAME).read_text().strip()
    }

    # create HuggingFaceModel with the image uri
    llm_model = HuggingFaceModel(role=role_arn,
                                 image_uri=experiment_config['image_uri'],
                                 env=model_config)

    print(f"Hugging face model defined using {model_config} -> {llm_model}")
    return llm_model

## Function to check the status of the endpoint
def check_endpoint_status(endpoint_name: str) -> str:
    resp = sm_client.describe_endpoint(EndpointName=endpoint_name)
    status = resp["EndpointStatus"]
    while status == "Creating":
        time.sleep(60)
        resp = sm_client.describe_endpoint(EndpointName=endpoint_name)
        status = resp["EndpointStatus"]
    return status

# Deploy the hugging face model
def deploy_hugging_face_model(experiment_config: Dict, llm_model: HuggingFaceModel) -> str:   
    tmout: int = experiment_config['env']['HEALTH_CHECK_TIMEOUT']
    llm = llm_model.deploy(initial_instance_count=experiment_config['env']['INSTANCE_COUNT'],
                           instance_type=experiment_config['instance_type'],
                           container_startup_health_check_timeout=tmout,
                           tags=tag)
    return llm.endpoint_name

# Function to deploy the model and create the endpoint
def deploy(experiment_config: Dict, role_arn: str) -> Dict[str, str]:
    logger.info("deploying the model using the llm_model and the configurations ....")


    print(f"Setting the model configurations .....")
    llm_model = create_hugging_face_model(experiment_config, role_arn)
    logger.info(f"the llm_model has been defined .... {llm_model}")

    llm_endpoint = deploy_hugging_face_model(experiment_config, llm_model)
    logger.info("Deploying the model now ....")

    status = check_endpoint_status(llm_endpoint)
    logger.info(f"Endpoint status: {status}")

    return dict(endpoint_name=llm_endpoint, 
                experiment_name=experiment_config['name'], 
                instance_type=experiment_config['instance_type'], 
                instance_count=experiment_config['instance_count'], 
                deployed=True)



import time
import fmbench
from typing import Dict
from fmbench.scripts import constants
from sagemaker.predictor import Predictor
from sagemaker.jumpstart.model import JumpStartModel


# Initialize the platform where this script deploys the model
PLATFORM: str = constants.PLATFORM_SAGEMAKER

tag = [
    {
        'Key': 'fmbench-version',
        'Value': fmbench.__version__
    }
]


def deploy(experiment_config: Dict, role_arn: str) -> Dict:
    model = JumpStartModel(
            model_id=experiment_config['model_id'],
            model_version=experiment_config['model_version'],
            image_uri=experiment_config['image_uri'],
            env=experiment_config['env'],
            role=role_arn,
            instance_type=experiment_config['instance_type']
        )

    sec, us = str(time.time()).split(".")
    ep_name = f"{experiment_config['ep_name']}-{sec}-{us}"
    accept_eula = experiment_config.get('accept_eula')
    if accept_eula is not None:
        predictor = model.deploy(initial_instance_count=experiment_config['instance_count'],
                                 accept_eula=accept_eula,
                                 endpoint_name=ep_name,
                                 tags=tag)
    else:
        predictor = model.deploy(initial_instance_count=experiment_config['instance_count'],
                                 endpoint_name=ep_name,
                                 tags=tag)

    return dict(endpoint_name=predictor.endpoint_name, 
                experiment_name=experiment_config['name'], 
                instance_type=experiment_config['instance_type'], 
                instance_count=experiment_config['instance_count'], 
                deployed=True)



import json
import os

import torch
from djl_python import Input, Output, PairList
from djl_python.transformers_neuronx import TransformersNeuronXService
from djl_python.test_model import decode_encoded_output_binary, create_concurrent_batch_request
import numpy as np
import time
import asyncio, threading, copy

import triton_python_backend_utils as pb_utils

_MODEL_ARGS_FILENAME: str = "model.json"

class TritonPythonModel:

  def initialize(self, args):
    self.logger = pb_utils.Logger
    self.model_config = json.loads(args["model_config"])
    text_output_config = pb_utils.get_output_config_by_name(self.model_config, "text_output")
    self.text_output_dtype = pb_utils.triton_string_to_numpy(text_output_config["data_type"])
    self.__tasks = set()
    self._init_service()
    self.__tasks_inited = False
      

  @staticmethod
  def auto_complete_config(auto_complete_model_config):
      
    inputs = [
    {"name": "text_input", "data_type": "TYPE_STRING", "dims": [1]},
    {
        "name": "sampling_parameters",
        "data_type": "TYPE_STRING",
        "dims": [1],
        "optional": True,
    }
    ]
    outputs = [{"name": "text_output", "data_type": "TYPE_STRING", "dims": [-1]}]

    config = auto_complete_model_config.as_dict()
    input_names = []
    output_names = []
    for input in config['input']:
      input_names.append(input['name'])
    for output in config['output']:
      output_names.append(output['name'])

    for input in inputs:
      if input['name'] not in input_names:
          auto_complete_model_config.add_input(input)
    for output in outputs:
      if output['name'] not in output_names:
          auto_complete_model_config.add_output(output)

    auto_complete_model_config.set_model_transaction_policy(dict(decoupled=True))
    auto_complete_model_config.set_max_batch_size(0)

    return auto_complete_model_config

  def _init_service(self):
    self.logger.log_info("Enter: _init_service")

    max_batch_size = int(self.model_config.get('max_batch_size', 0))
    assert (max_batch_size == 0), "max_batch_size must be 0"
    self.using_decoupled = pb_utils.using_decoupled_model_transaction_policy(self.model_config) 
    assert (
        self.using_decoupled 
    ), "Python backend must be configured to use decoupled model transaction policy"

    model_args_filepath = os.path.join( 
        pb_utils.get_model_dir(), _MODEL_ARGS_FILENAME
    )
    assert os.path.isfile(
        model_args_filepath
    ), f"'{_MODEL_ARGS_FILENAME}' containing model args must be provided in '{pb_utils.get_model_dir()}'"
    with open(model_args_filepath) as file:
      self.model_args = json.load(file)

    self.batch_size = self.model_args.get("batch_size", None)
    if self.batch_size is None:
      self.batch_size = self.model_args.get("max_rolling_batch_size", 1)

    self.logger.log_info(f"initialize service: {self.model_args}")
    self.__service = TransformersNeuronXService()
    self.__service.initialize(self.model_args)
    self.logger.log_info("service initialized.")

    self.logger.log_info("Create request asyncio queue: maxsize {self.batch_size}")
    self.__request_queue = asyncio.Queue(maxsize=self.batch_size)

    self.logger.log_info("Create response asyncio queue: maxsize {self.batch_size}")
    self.__response_queue = asyncio.Queue(maxsize=self.batch_size)

    self.logger.log_info("Exit: _init_service")

  def get_sampling_params_dict(self, params_json):              
    params_dict = json.loads(params_json) if params_json else {}

    float_keys = [
        "temperature",
        "top_p"
    ]
    for k in float_keys:
        if k in params_dict:
            params_dict[k] = float(params_dict[k])
        
    int_keys = ["sequence_length", "top_k"]
    for k in int_keys:
        if k in params_dict:
            params_dict[k] = int(params_dict[k])

    if not params_dict:
        params_dict["max_new_tokens"] = 8192
        params_dict["top_k"] = 50
    else:
        if "max_new_tokens" not in params_dict:
          params_dict["max_new_tokens"] = 8192

    return params_dict

  async def __init_tasks(self):
    self.logger.log_info("Start respond loop")
    task = asyncio.create_task(self.__respond_loop())
    self.__tasks.add(task)
    task.add_done_callback(self.__tasks.discard)
    
    self.logger.log_info("Start generate loop")
    task = asyncio.create_task(self.__generate_loop())
    self.__tasks.add(task)
    task.add_done_callback(self.__tasks.discard)

    self.__tasks_inited = True

  async def execute(self, requests):
    if not self.__tasks_inited:
      try:
        await self.__init_tasks()
      except KeyError:
        print("Future not found or has already completed.")

    for request in requests:
      try:
        await self.__request_queue.put(request)
      except KeyError:
        print("Future not found or has already completed.")

  async def __check_new_requests(self):
    self.__new_requests = []
    if len(self.__requests) == 0:
      new_request = await self.__request_queue.get()
      self.__request_queue.task_done()
      self.__new_requests.append(new_request)
      self.__requests.append(new_request)

    while len(self.__requests) < self.batch_size:
      try:
        await asyncio.sleep(.001)
        new_request = self.__request_queue.get_nowait()
        self.__request_queue.task_done()
        self.__requests.append(new_request)
        self.__new_requests.append(new_request)
      except asyncio.QueueEmpty:
        break

  def __inference(self):

    for request in self.__new_requests:
      prompts = pb_utils.get_input_tensor_by_name(request, "text_input").as_numpy().flatten().tolist()
      inputs = [ p.decode("utf-8") if isinstance(p, bytes) else p for p in prompts]
      inputs = inputs[0] if len(inputs) == 1 else inputs

      parameters_input_tensor = pb_utils.get_input_tensor_by_name(request, "sampling_parameters")
      if parameters_input_tensor:
        parameters = parameters_input_tensor.as_numpy().flatten()
        parameters = parameters.tolist()[0] # assume uniform sampling parameters in batch
        parameters = parameters.decode('utf-8') if isinstance(parameters, bytes) else parameters
      else:
        parameters = request.parameters()
      params = self.get_sampling_params_dict(parameters)
      
      self.__input_list.append({"inputs": inputs,"parameters": params})

    n_new_requests = len(self.__new_requests)
    if n_new_requests > 0:
      properties = [{"eula": "true", "Content-type": "application/json"}]*len(self.__input_list)
      input = create_concurrent_batch_request(inputs=self.__input_list,
                                    properties=properties,
                                    serving_properties=copy.deepcopy(self.model_args))
      output = self.__service.inference(input)
      for i in range(output.content.size()):
        result = decode_encoded_output_binary(output.content.value_at(i))
        if i < len(self.__results):
          self.__results[i] = result
        else:
          self.__results.append(result)
    elif self.__service.rolling_batch: 
      self.__results = self.__service.rolling_batch.inference([])
    
  async def __generate_loop(self):
    
  
    while True:
      try:
        await self.__check_new_requests()

        unique_reqs = { f"{x}" for x in self.__requests }
        assert len(unique_reqs) == len(self.__requests), \
          f"requests are not unique, {self.__requests}"

        self.__inference()

        assert len(self.__requests) == len(self.__input_list), \
          f"requests: {len(self.__requests)} != input_list: {len(self.__input_list)}"

        assert len(self.__requests) == len(self.__results), \
          f"requests: {len(self.__requests)} != results: {len(self.__results)}"

        finished = []
        for i in range(len(self.__requests)):
          res = self.__results[i]
          if str(res['last']).lower() == 'true':
            req = self.__requests[i]
            input = self.__input_list[i]
            try:
              self.__response_queue.put_nowait((req, res))
            except asyncio.QueueFull:
              self.logger.log_info("response queue is full; await put")
              await self.__response_queue.put((req, res))
            finished.append((req, res, input))
          
        for item in finished:
          req, res, input = item
          self.__requests.remove(req)
          self.__results.remove(res)
          self.__input_list.remove(input)

        assert len(self.__requests) == len(self.__input_list), \
          f"requests: {len(self.__requests)} != input_list: {len(self.__input_list)}"

        assert len(self.__requests) == len(self.__results), \
          f"requests: {len(self.__requests)} != results: {len(self.__results)}"

      except Exception as e:
        self.logger.log_error(f"Unpexpected error: {e}. Inflight requests discarded. Reset engine.")
        self.reset()

  def reset(self):
    self.__requests = []
    self.__results = []
    self.__input_list = []

    if self.__service.rolling_batch:
      self.__service.rolling_batch.reset()

  async def __respond_loop(self):
    self.reset()

    while True:
      try:
        req, res = await self.__response_queue.get()
        self.__response_queue.task_done()
        
        t = threading.Thread(target=self.__send_response, 
          kwargs={"request": req, "response": res})
        t.start()
      except Exception as e:
        print(f"Error: respond loop exception {e}")

  def __send_response(self, request, response: dict):
    try:
      response_sender = request.get_response_sender()
      text_output = response['data']
    
      out_tensor = pb_utils.Tensor("text_output", np.array(text_output).astype(self.text_output_dtype))
      inference_response = pb_utils.InferenceResponse(output_tensors=[out_tensor])

      response_sender.send(inference_response, flags=pb_utils.TRITONSERVER_RESPONSE_COMPLETE_FINAL)
    except Exception as e:
      print(f"send error: {request}", flush=True)
      print(e, flush=True)

  def finalize(self):
    self.logger.log_info("Cleaning up...")


import os
import json
import copy
import time
import stat
import logging
import requests
import tempfile
import subprocess
import pandas as pd
from pathlib import Path
from datetime import datetime
from typing import Dict, Optional
from fmbench.scripts import constants
from fmbench.utils import count_tokens
from fmbench.scripts.ec2_metrics import collect_ec2_metrics, stop_collect
from fmbench.scripts.inference_containers.utils import get_accelerator_type
from fmbench.scripts.fmbench_predictor import (FMBenchPredictor,
                                               FMBenchPredictionResponse)
from fmbench.scripts.inference_containers.utils import (STOP_AND_RM_CONTAINER,
                                                        FMBENCH_MODEL_CONTAINER_NAME)
                                            
# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class EC2Predictor(FMBenchPredictor):
    # overriding abstract method
    def __init__(self,
                 endpoint_name: str,
                 inference_spec: Optional[Dict], 
                 metadata: Optional[Dict]):
        try:
            self._endpoint_name: str = endpoint_name
            self._inference_spec: Dict = inference_spec
            self._accelerator = get_accelerator_type()
            # Start collecting EC2 metrics. This will be called once.
            collect_ec2_metrics()
        except Exception as e:
            logger.error(f"create_predictor, exception occured while creating predictor "
                         f"for endpoint_name={self._endpoint_name}, exception={e}")
        logger.info(f"_endpoint_name={self._endpoint_name}, _inference_spec={self._inference_spec}")

    def get_prediction(self, payload: Dict) -> FMBenchPredictionResponse:
        response_json: Optional[Dict] = None
        response: Optional[str] = None
        latency: Optional[float] = None
        TTFT: Optional[float] = None
        TPOT: Optional[float] = None
        TTLT: Optional[float] = None
        prompt_tokens: Optional[int] = None
        completion_tokens: Optional[int] = None
        container_type: Optional[str] = None
        # get the prompt for the EKS endpoint
        prompt: str = payload['inputs']
        # represents the number of tokens in the prompt payload
        prompt_tokens = count_tokens(payload['inputs'])
        try:            
            split_input_and_inference_params: Optional[bool] = None
            if self._inference_spec is not None:
                split_input_and_inference_params = self._inference_spec.get("split_input_and_parameters")
                container_type = self._inference_spec.get("container_type", constants.CONTAINER_TYPE_DJL)
                logger.debug(f"split input parameters is: {split_input_and_inference_params}, "
                             f"container_type={container_type}")
            # this is the POST request to the endpoint url for invocations that 
            # is given to you as you deploy a model on EC2 using the DJL serving stack
            if container_type == constants.CONTAINER_TYPE_DJL:
                payload = payload | dict(parameters=self._inference_spec["parameters"])
                st = time.perf_counter()
                response = requests.post(self._endpoint_name, json=payload)
                # record the latency for the response generated
                latency = time.perf_counter() - st                
                #logger.info(f"full_output={response.text}")
                response.raise_for_status()
                """
                the output is of the form
                {"generated_text": "\n\nSuining had a population of 658,798 in 2002."}
                we only need the generated_text field from this
                """
                full_output = json.loads(response.text).get("generated_text")
                if full_output is None:
                    logger.error(f"failed to extract output from response text, response text = \"{response.text}\"")      
            elif container_type == constants.CONTAINER_TYPE_TRITON:
                if self._accelerator == constants.ACCELERATOR_TYPE.NEURON:
                    triton_payload = dict(text_input=payload["inputs"],
                                          sampling_parameters=json.dumps(self._inference_spec["parameters"]))
                else:
                    triton_payload = dict(text_input=payload["inputs"]) | self._inference_spec["parameters"]

                logger.info(f"Endpoint name is: {self._endpoint_name}, triton payload is: {triton_payload}")
                st = time.perf_counter()
                response = requests.post(self._endpoint_name, json=triton_payload)
                # record the latency for the response generated
                latency = time.perf_counter() - st
                response.raise_for_status()
                response_json = json.loads(response.text)
                full_output = response_json['text_output']
            elif container_type == constants.CONTAINER_TYPE_VLLM:
                # vllm uses prompt rather than input and then
                # the code in the calling function still expects input
                # so make a copy
                payload2 = copy.deepcopy(payload)
                payload2['prompt'] = payload2.pop('inputs')
                payload2 = payload2 | self._inference_spec["parameters"]
                st = time.perf_counter()
                response = requests.post(self._endpoint_name, json=payload2)
                # record the latency for the response generated
                latency = time.perf_counter() - st
                response.raise_for_status()
                """
                the output is of the form
                {
                    "id": "cmpl-e9d590543c374d828d724a228fae2604",
                    "object": "text_completion",
                    "created": 1731734600,
                    "model": "meta-llama/Llama-3.2-1b-Instruct",
                    "choices": [
                        {
                            "index": 0,
                            "text": "\n\nThey are both tennis players.",
                            "logprobs": null,
                            "finish_reason": "stop",
                            "stop_reason": null,
                            "prompt_logprobs": null
                        }
                    ],
                    "usage": {
                        "prompt_tokens": 1392,
                        "total_tokens": 1400,
                        "completion_tokens": 8,
                        "prompt_tokens_details": null
                    }
                }
                we only need the response field from this
                """
                full_output = None
                choices = json.loads(response.text).get("choices")
                if choices is not None:
                    full_output = choices[0].get("text")
                if full_output is None:
                    logger.error(f"failed to extract output from response text, response text = \"{response.text}\"")
            elif container_type == constants.CONTAINER_TYPE_OLLAMA:
                # ollama uses prompt rather than input and then
                # the code in the calling function still expects input
                # so make a copy
                payload2 = copy.deepcopy(payload)
                payload2['prompt'] = payload2.pop('inputs')
                payload2 = payload2 | self._inference_spec["parameters"]
                st = time.perf_counter()
                response = requests.post(self._endpoint_name, json=payload2)
                # record the latency for the response generated
                latency = time.perf_counter() - st
                response.raise_for_status()
                """
                the output is of the form
                {"model":"llama3.1:8b",
                 "created_at":"2024-10-27T13:44:12.400070826Z",
                 "response":"Once upon a time, in a small village nestled",
                 "done":true,
                 "done_reason":"length",
                 "context":[128006,882,128007,271,73457,757,264,3446,128009,128006,78191,128007,271,12805,5304,264,892,11,304,264,2678,14458,89777],
                 "total_duration":195361027,
                 "load_duration":25814720,
                 "prompt_eval_count":14,
                 "prompt_eval_duration":14601000,
                 "eval_count":10,
                 "eval_duration":96146000}
                we only need the response field from this
                """
                full_output = json.loads(response.text).get("response")
                if full_output is None:
                    logger.error(f"failed to extract output from response text, response text = \"{response.text}\"")              
            else:
                raise ValueError("container_type={container_type}, dont know how to handle this") 

            answer_only = full_output.replace(prompt, "", 1).strip('["]?\n')
            response_json = dict(generated_text=answer_only)
            # counts the completion tokens for the model using the default/user provided tokenizer
            completion_tokens = count_tokens(response_json.get("generated_text"))
        except Exception as e:
            logger.error(f"get_prediction, exception occurred while getting prediction for payload={payload} "
                         f"from predictor={self._endpoint_name}, response={response}, exception={e}")

        return FMBenchPredictionResponse(response_json=response_json,
                                         latency=latency,
                                         time_to_first_token=TTFT,
                                         time_per_output_token=TPOT,
                                         time_to_last_token=TTLT,
                                         completion_tokens=completion_tokens,
                                         prompt_tokens=prompt_tokens)

    def shutdown(self) -> None:
        """Represents the function to shutdown the predictor
           cleanup the endpooint/container/other resources
        """
        # Stop collecting EC2 metrics either when the model container is stopped and removed, 
        # or once the benchmarking test has completed 
        stop_collect()

        script = f"""#!/bin/sh

        {STOP_AND_RM_CONTAINER}
        """
        tmpdir = tempfile.gettempdir()
        script_file_path = os.path.join(tmpdir, "shutdown_container.sh")
        Path(script_file_path).write_text(script)
        st = os.stat(script_file_path)
        os.chmod(script_file_path, st.st_mode | stat.S_IEXEC)

        logger.info(f"going to run script {script_file_path}")
        subprocess.run(["bash", script_file_path], check=True)
        logger.info(f"done running bash script")
        return None
    
    @property
    def endpoint_name(self) -> str:
        """The endpoint name property."""
        return self._endpoint_name

    # The rest ep is deployed on an instance that incurs hourly cost hence, the calculcate cost function
    # computes the cost of the experiment on an hourly basis. If your instance has a different pricing structure
    # modify this function.
    def calculate_cost(self,
                       instance_type: str,
                       instance_count: int,
                       pricing: Dict,
                       duration: float,
                       prompt_tokens: int,
                       completion_tokens: int) -> float:
        """Calculate the cost of each experiment run."""
        experiment_cost: Optional[float] = None
        try:
            instance_based_pricing = pricing['pricing']['instance_based']
            hourly_rate = instance_based_pricing.get(instance_type, None)
            logger.info(f"the hourly rate for running on {instance_type} is {hourly_rate}, instance_count={instance_count}")
            # calculating the experiment cost for instance based pricing
            instance_count = instance_count if instance_count else 1
            experiment_cost = (hourly_rate / 3600) * duration * instance_count
        except Exception as e:
            logger.error(f"exception occurred during experiment cost calculation, exception={e}")
        return experiment_cost
    
    def get_metrics(self,
                start_time: datetime,
                end_time: datetime,
                period: int = 60) -> pd.DataFrame:
        """
        Retrieves EC2 system metrics from the CSV file generated during metric collection.

        Args:
            start_time (datetime): Start time of metrics collection
            end_time (datetime): End time of metrics collection
            period (int): Sampling period in seconds (default 60)
        
        Returns:
            pd.DataFrame: DataFrame containing system metrics
        """
        try:
            filtered_metrics: Optional[pd.DataFrame] = None
            # Read the CSV file generated by collect_ec2_metrics()
            metrics_df = pd.read_csv(constants.EC2_SYSTEM_METRICS_FNAME, parse_dates=['timestamp'])
            filtered_metrics = metrics_df[(metrics_df['timestamp'] >= start_time) & 
                                            (metrics_df['timestamp'] <= end_time)]
        except FileNotFoundError:
            logger.warning("Metrics CSV file containin the EC2 metrics not found.")
            filtered_metrics = None
        except Exception as e:
            logger.error(f"Error retrieving metrics: {e}")
        return filtered_metrics

    @property
    def inference_parameters(self) -> Dict:
        """The inference parameters property."""
        return self._inference_spec.get("parameters")
    
    @property
    def platform_type(self) -> Dict:
        """The inference parameters property."""
        return constants.PLATFORM_EC2


def create_predictor(endpoint_name: str, inference_spec: Optional[Dict], metadata: Optional[Dict]):
    return EC2Predictor(endpoint_name, inference_spec, metadata)



import importlib.metadata

try:
    __version__ = importlib.metadata.version("fmbench")
except importlib.metadata.PackageNotFoundError:
    # Package is not installed, fallback to a default version
    __version__ = "None"



import os
import json
import math
import time
import boto3
import logging
import requests
import datetime
import pandas as pd
from datetime import datetime
from fmbench.scripts import constants
from fmbench.utils import count_tokens
from typing import Dict, Optional, List
from fmbench.scripts.fmbench_predictor import (FMBenchPredictor,
                                               FMBenchPredictionResponse)

# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class EKSPredictor(FMBenchPredictor):
    # overriding abstract method
    def __init__(self,
                 endpoint_name: str,
                 inference_spec: Optional[Dict],
                 metadata: Optional[Dict]):
        try:
            self._endpoint_name: str = endpoint_name
            self._inference_spec: Dict = inference_spec 
        except Exception as e:
            logger.error(f"create_predictor, exception occured while creating predictor "
                         f"for endpoint_name={self._endpoint_name}, exception={e}")
        logger.info(f"_endpoint_name={self._endpoint_name}, _inference_spec={self._inference_spec}")

    def get_prediction(self, payload: Dict) -> FMBenchPredictionResponse:
        response_json: Optional[Dict] = None
        response: Optional[str] = None
        latency: Optional[float] = None
        TTFT: Optional[float] = None
        TPOT: Optional[float] = None
        TTLT: Optional[float] = None
        prompt_tokens: Optional[int] = None
        completion_tokens: Optional[int] = None
        timeout: Optional[int] = None
        auth: Optional[Dict] = None
        # get the prompt for the EKS endpoint
        prompt: str = payload['inputs']
        # represents the number of tokens in the prompt payload
        prompt_tokens = count_tokens(payload["inputs"])
        try:
            st = time.perf_counter()
            split_input_and_inference_params: Optional[bool] = None
            if self._inference_spec is not None:
                split_input_and_inference_params = self._inference_spec.get("split_input_and_parameters")
                logger.info(f"split input parameters is: {split_input_and_inference_params}")
                timeout = self._inference_spec.get("timeout", 180)
                auth = self._inference_spec.get("auth", None)
                logger.info(f"Initializing the timeout to: {timeout}, auth to configured authentication information")
                # Use the parameters that the model needs at inference. 
                # parameters: Optional[Dict] = inference_spec.get('parameters')

            # This endpoint only supports the GET method now, you can add support for POST method if your endpoint supports it.
            # As an example, the following URL is used with a query added at the end of the URL.
            # http://<NLB_DNS_NAME>/serve/infer?sentence=<prompt_payload>
            response = requests.get(f"{self._endpoint_name}/{prompt}",timeout=timeout) 

            latency = time.perf_counter() - st
            logger.info(response)
            # the response from the model on ray serve from the url prompt is given in this format. 
            # For other response types, change the logic below and add the response in the `generated_text` key within the response_json dict
            response.raise_for_status()
            full_output = response.text
            answer_only = full_output.replace(prompt, "", 1).strip('["]?\n')
            response_json = dict(generated_text=answer_only)
            # counts the completion tokens for the model using the default/user provided tokenizer
            completion_tokens = count_tokens(response_json.get("generated_text"))
        except requests.exceptions.RequestException as e:
            logger.error(f"get_prediction, exception occurred while getting prediction for payload={payload} "
                         f"from predictor={self._endpoint_name}, response={response}, exception={e}")
        return FMBenchPredictionResponse(response_json=response_json,
                                         latency=latency,
                                         time_to_first_token=TTFT,
                                         time_per_output_token=TPOT,
                                         time_to_last_token=TTLT,
                                         completion_tokens=completion_tokens,
                                         prompt_tokens=prompt_tokens)    

    @property
    def endpoint_name(self) -> str:
        """The endpoint name property."""
        return self._endpoint_name

    def get_metrics(self,
                    start_time: datetime,
                    end_time: datetime,
                    period: int = 60) -> pd.DataFrame:
        # not implemented
        return None

    # The rest ep is deployed on an instance that incurs hourly cost hence, the calculcate cost function
    # computes the cost of the experiment on an hourly basis. If your instance has a different pricing structure
    # modify this function.
    def calculate_cost(self,
                       instance_type: str,
                       instance_count: int,
                       pricing: Dict,
                       duration: float,
                       prompt_tokens: int,
                       completion_tokens: int) -> float:
        """Calculate the cost of each experiment run."""
        experiment_cost: Optional[float] = None
        try:
            instance_based_pricing = pricing['pricing']['instance_based']
            hourly_rate = instance_based_pricing.get(instance_type, None)
            logger.info(f"the hourly rate for running on {instance_type} is {hourly_rate}, instance_count={instance_count}")
            # calculating the experiment cost for instance based pricing
            instance_count = instance_count if instance_count else 1
            experiment_cost = (hourly_rate / 3600) * duration * instance_count
        except Exception as e:
            logger.error(f"exception occurred during experiment cost calculation, exception={e}")
        return experiment_cost

    def shutdown(self) -> None:
        """Represents the function to shutdown the predictor
           cleanup the endpooint/container/other resources
        """
        return None
    
    @property
    def inference_parameters(self) -> Dict:
        """The inference parameters property."""
        return self._inference_spec.get("parameters")

    @property
    def platform_type(self) -> Dict:
        """The inference parameters property."""
        return constants.PLATFORM_EC2
    
def create_predictor(endpoint_name: str, inference_spec: Optional[Dict], metadata: Optional[Dict]):
    return EKSPredictor(endpoint_name, inference_spec, metadata)



"""
Deploys a model from HuggingFace on Amazon EC2

1. Configuration is read from the configured serving.properties file.
2. A hf_token.txt file is required to download the model from Hugging Face.
"""
# Import necessary libraries
import os
import re
import sys
import time
import json
import stat
import docker
import inspect
import logging
import requests
import tempfile
import subprocess
from pathlib import Path
from typing import Dict, Union
from fmbench.scripts import constants
from ec2_metadata import ec2_metadata
from fmbench.scripts.inference_containers import (djl, vllm, triton)
from fmbench.scripts.constants import (IS_NEURON_INSTANCE, LISTEN_PORT)
from fmbench.scripts.prepare_for_multi_model_containers import prepare_docker_compose_yml

# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Initialize the platform where this script deploys the model
PLATFORM: str = constants.PLATFORM_EC2

# globals
HF_TOKEN_FNAME: str = os.path.join(os.path.dirname(os.path.realpath(__file__)), "hf_token.txt")
FMBENCH_MODEL_CONTAINER_NAME: str = "fmbench_model_container"

def _set_up(model_name: str, local_model_path: str):
    """
    Create the model serving.properties file locally in a model directory
    """
    # make the model directory with serving.properties
    directory = os.path.join(local_model_path, model_name)
    os.makedirs(directory, exist_ok=True)
    logger.info(f"Creating the local directory: {directory}")
    
    # return the directory we created
    return directory

def _create_deployment_script(image_uri,
                              container_type,
                              privileged_mode,
                              region,
                              model_name,
                              model_id,
                              hf_token,
                              directory,
                              model_loading_timeout,
                              env,
                              model_copies,
                              is_neuron_instance):
    """
    Write a deployment script for model container
    """
    logger.info(inspect.getargvalues(inspect.currentframe()))
    privileged_str: str = "--privileged" if privileged_mode else ""
    env_str: str = ""
    if env is not None:
        for k, v in env.items():
            env_str += f"-e {k}={v} "

    match container_type:
        case constants.CONTAINER_TYPE_DJL:
            deploy_script_content = djl.create_script(region, image_uri, model_id, model_name, env_str, privileged_str, hf_token, directory)
        case constants.CONTAINER_TYPE_VLLM:
            deploy_script_content = vllm.create_script(region, image_uri, model_id, model_name, env_str, privileged_str, hf_token, directory)
        case constants.CONTAINER_TYPE_TRITON:
            deploy_script_content = triton.create_script(region, image_uri, model_id, model_name, env_str, privileged_str, hf_token, directory)
        case _:
            raise ValueError(f"dont know how to handle container_type={container_type}")
    script_file_path = os.path.join(directory, "deploy_model.sh")
    Path(script_file_path).write_text(deploy_script_content)
    st = os.stat(script_file_path)
    os.chmod(script_file_path, st.st_mode | stat.S_IEXEC)

    logger.info(f"deploy_model.sh content: {deploy_script_content}")
    logger.info(f"The deploy_model.sh file has been created in {script_file_path}")
    return script_file_path

def _run_container(script_file_path):
    """
    Runs the deploy_model.sh bash script with the provided arguments.
    """
    logger.info(f"Running container script at {script_file_path}")
    # Create a Docker client
    client = docker.from_env()

    try:        
        logger.info(f"going to run script {script_file_path}")
        subprocess.run(["bash", script_file_path], check=True)
        logger.info(f"done running bash script")
        return True    
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
    return False

def _check_model_deployment(endpoint, model_id, container_type, model_loading_timeout, model_copies, is_neuron_instance):
    """
    Check the model deployment status and wait for the model to be ready.
    """
    start_time = time.time()
    logger.info(f"Checking deployment status at {endpoint}")
    if container_type == constants.CONTAINER_TYPE_DJL:
        data = {"inputs": "tell me a story of the little red riding hood"}
    elif container_type == constants.CONTAINER_TYPE_VLLM:
        data = {"model": model_id,  # Specify the model to use
                "prompt": "tell me a story of the little red riding hood",}
    elif container_type == constants.CONTAINER_TYPE_TRITON:
        if is_neuron_instance:
            data = {
                    "text_input": "tell me a story of the little red riding hood",
                    "sampling_parameters": json.dumps({"top_k": 50})
                }
        else:
            data = {"text_input": "tell me a story of the little red riding hood", "max_tokens": 50}
    headers = {"content-type": "application/json"}
    container_check_timeout = 60
    logger.info(f"going to check every {container_check_timeout}s for the inference endpoint to be up...")
    logger.info(f"this will take at least 10 minutes or so, please be patient...")
    while time.time() - start_time < model_loading_timeout:        
        try:
            response = requests.post(endpoint, headers=headers, json=data)
            logger.info(f"response is: {response.text}")
            if response.status_code == 200:
                logger.info("model deployment successful!")

                # if model_copies != 1 then wait for some more time to give
                # all the containers a chance to be up
                if model_copies > 1:
                    # this extra wait is not needed if this is a triton server running on a gpu
                    # because we start multiple model servers in a single container
                    if is_neuron_instance is False and container_type == constants.CONTAINER_TYPE_TRITON:
                        logger.info(f"this is a triton container running on a gpu instance, skipping the additional wait")
                    else:
                        additional_sleep_time = model_copies * container_check_timeout
                        logger.info(f"since model_copies={model_copies}, waiting for an addition {additional_sleep_time}s "
                                    f"to allow all model endpoints to come up")
                        time.sleep(additional_sleep_time)
                total_wait_time = time.time() - start_time
                logger.info("marking endpoint in service, total_wait_time={total_wait_time}s")
                return "InService"
            else:
                logger.info(f"model deployment is not ready yet. Return code: {response.status_code}")
                logger.info(f"waiting for {container_check_timeout}s before retrying")
                time.sleep(container_check_timeout)
        except Exception as e:
            logger.error(f"error occurred while deploying the endpoint: {e}")
            logger.info(f"waiting for {container_check_timeout}s before retrying")
            time.sleep(container_check_timeout)
    return "Failed"

def deploy(experiment_config: Dict, role_arn: str) -> Dict:
    """
    Function to deploy the model and create the endpoint
    """
    image_uri: str = experiment_config['image_uri']
    ep_name: str = experiment_config['ep_name']
    model_id: str = experiment_config['model_id']
    model_name: str = Path(model_id).name
    logger.info(f"Going to deploy model: {model_name}")

    ep_name: str = experiment_config['ep_name']
    # extract port number from the endpoint name
    # 8080 from "http://localhost:8080/v2/models/ensemble/generate"
    # this is the port on which either the inference container (if there
    # is a single container) or the load balancer in case of multiple containers
    # will listen for new connnections
    match = re.search(r"http://[^:/]+:(\d+)", ep_name)
    listen_port = int(match.group(1)) if match else constants.LISTEN_PORT
    logger.info(f"ep_name={ep_name}, listen_port={listen_port}")
    region: str = experiment_config['region']
    privileged_mode: str = experiment_config['ec2'].get('privileged_mode', False)
    container_type: str = experiment_config['inference_spec'].get('container_type', constants.CONTAINER_TYPE_DJL)
    env = experiment_config.get('env')
    container_params: str = experiment_config['inference_spec'].get('container_params', None)
    if container_params is None:
        serving_properties = experiment_config.get('serving.properties', None)
        logger.info(f"Serving properties found in experiment config: {serving_properties}")
    else: 
        serving_properties = container_params.get('serving.properties', None)
        logger.info(f"Serving properties found in container params in the config: {serving_properties}")
    logger.info(f"Serving properties provided to deploy on container type={container_type} using image_uri={image_uri}: {serving_properties}")
    model_loading_timeout: int = experiment_config['ec2']['model_loading_timeout']
    
    HF_TOKEN: str = Path(HF_TOKEN_FNAME).read_text().strip()
    dir_path = home_dir = os.getenv("HOME", str(Path.home()))
    logger.info(f"Home directory: {dir_path}")
    
    # print the ec2 instance details for it to get logged
    logger.info(f"EC2 instance type: {ec2_metadata.instance_type}, AMI ID: {ec2_metadata.ami_id}")
    
    # HF token is mandatory even if no env vars were specific in the config file    
    if env is None:
        env = {"HF_TOKEN": HF_TOKEN}
    else:
        env["HF_TOKEN"] = HF_TOKEN
    logger.info(f"env={env}")    

    model_directory = _set_up(model_name, dir_path)
    is_neuron_instance = IS_NEURON_INSTANCE(experiment_config['instance_type'])
    model_copies = experiment_config['inference_spec'].get('model_copies', '1')


    # if this is a neuron instance and we are using the djl serving inference container
    # then use the docker compose approach so we first create the docker compose file
    # and then create the deployment script, otherwise we create the deployment script
    # directly (see call to _create_deployment_script below)
    model_copies_actual = 1
    if (container_type == constants.CONTAINER_TYPE_DJL) or (container_type == constants.CONTAINER_TYPE_TRITON):  
        logger.info(f"container_type={container_type}, is_neuron_instance={is_neuron_instance}, going to create docker compose yml")
        model_copies_actual = prepare_docker_compose_yml(model_id=model_id,
                                                         model_copies=model_copies,
                                                         inference_params=experiment_config['inference_spec'],
                                                         image=image_uri,
                                                         user=container_type,
                                                         shm_size=experiment_config['inference_spec']['shm_size'],
                                                         env=env,
                                                         serving_properties=serving_properties,
                                                         dir_path=dir_path,
                                                         listen_port=listen_port)   
    logger.info("Creating the deployment script in model directory")
    deployment_script_path = _create_deployment_script(image_uri,
                                                       container_type,
                                                       privileged_mode,
                                                       region,
                                                       model_name,
                                                       model_id,
                                                       HF_TOKEN,
                                                       model_directory,
                                                       model_loading_timeout,
                                                       env,
                                                       model_copies,
                                                       is_neuron_instance)

    logger.info("Running the deployment script")
    ran_container = _run_container(deployment_script_path)

    # initialize with None values for error case
    deployment_result: Dict = dict(endpoint_name=None, 
                                   experiment_name=None,
                                   instance_type=None,
                                   instance_count=None, 
                                   deployed=False)
    if ran_container:
        logger.info("Container ran successfully")
        ep_status = _check_model_deployment(ep_name, model_id, container_type, model_loading_timeout, model_copies_actual, is_neuron_instance)
        logger.info(f"Endpoint status: {ep_status}")
        if ep_status == "InService":
            logger.info("Model endpoint running!")
            deployment_result['endpoint_name'] = ep_name
            deployment_result['experiment_name'] = experiment_config['name']
            deployment_result['instance_type'] = experiment_config['instance_type']
            deployment_result['instance_count'] = experiment_config['instance_count']
            deployment_result['deployed'] = True
            return deployment_result
        elif ep_status == "Failed":
            logger.error("Model endpoint not running!")
            return deployment_result
    else:
        logger.error("Container did not run successfully")
        return deployment_result



"""
Triton specific code
"""
import os
import json
import stat
import shutil
import logging
from pathlib import Path
from typing import Tuple, List
import fmbench.scripts.constants as constants
from typing import Dict, List, Optional, Tuple, Union
from fmbench.scripts.inference_containers.utils import (STOP_AND_RM_CONTAINER,
                                                        FMBENCH_MODEL_CONTAINER_NAME)

logging.basicConfig(format='[%(asctime)s] p%(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

CONFIG_PROPERTIES: str = """
            inference_address=http://0.0.0.0:{port}
            management_address=http://0.0.0.0:{port}
            cluster_address=http://0.0.0.0:8888
            model_store=/opt/ml/model
            load_models=ALL
            """

def create_script(region, image_uri, model_id, model_name, env_str, privileged_str, hf_token, directory):
    """
    Script for running the docker container for the inference server
    """
    script = f"""#!/bin/sh
        echo "Going to download model container"
        echo "Content in docker command: {region}, {image_uri}, {model_name}"
        
        # using the locally built triton image

        # Run the new Docker container with specified settings
        cd {directory}
        # shutdown existing docker compose        
        docker compose down
        {STOP_AND_RM_CONTAINER}

        # download the model
        export HF_TOKEN={hf_token}
        pip freeze | grep huggingface-hub
        RESULT=$?
        if [ $RESULT -eq 0 ]; then
          echo huggingface-hub is already installed..
        else
          echo huggingface-hub is not installed, going to install it now
          pip install -U "huggingface_hub[cli]"
        fi

        huggingface-cli download {model_id} --local-dir $HOME/{model_id}

        # bring up the inference container and load balancer
        docker compose up -d
        cd -
        echo "started docker compose in daemon mode"
    """
    return script


def handle_triton_serving_properties_and_inf_params(triton_dir: str,
                                                    tp_degree: int,
                                                    container_params: Dict,
                                                    hf_model_id: str, 
                                                    backend: constants.BACKEND):
    """
    Substitutes parameters within the triton model repository files for different container types
    For both containers (djl and vllm), the only file that is created based on custom user provided
    metrics is the "model.json" file. Other files (config.pbtxt and model.py) remain unchanged. The
    model.json file contains the model id, tp degree, and another set of container parameters provided
    in the configuration file specific to the djl/vllm containers
    """
    try:
        model_json_file: str = "model.json"
        model_json_content: Dict = {}
        model_json_path: str = os.path.join(triton_dir, model_json_file)
        if container_params is None or container_params == {}:
            logger.error(f"container_params is not specified, fix the config file to proceed")
            ValueError("container_params is not specified, fix the config file to proceed")
        # if there is a tp degree in inference container parameters, delete it since those will already be added here. TP
        # degree is required in the container_params within the configuration file        
        del container_params['tp_degree']
        # if serving.properties are provided in the inference container parameters, then pop 
        # it out, since it is not needed for deploying the model using triton on neuron
        if 'serving.properties' in container_params:
            del container_params['serving.properties']
            logger.info(f"Removed the serving properties for deploying the model on triton on neuron: {container_params}")
        else:
            logger.info(f"No serving properties provided, using the inference model parameters directly in model deployment")
        if backend == constants.BACKEND.VLLM_BACKEND:
            logger.info(f"Handling parameter subsitution for {backend}")
            with open(model_json_path, "w") as f:
                model_json_content['model'] = hf_model_id
                model_json_content["tensor_parallel_size"] = tp_degree
                # update the model.json to contain additional variables, such as
                # max_num_seqs, max_model_len, batch_size and more
                model_json_content.update(container_params)
                json.dump(model_json_content, f, indent=2)
                logger.info(f"Updated {model_json_path}, model.json={model_json_content}")
        elif backend == constants.BACKEND.DJL_BACKEND:
            logger.info(f"Handling parameter subsitution for {backend}")
            with open(model_json_path, "w") as f:
                model_json_content['model_id'] = hf_model_id
                model_json_content["tensor_parallel_degree"] = tp_degree
                # update the model.json to contain additional variables, such as
                # max_num_seqs, max_model_len, batch_size and more
                model_json_content.update(container_params)
                json.dump(model_json_content, f, indent=2)
                logger.info(f"Updated {model_json_path}, model.json={model_json_content}")
        else:
            # If there are no backend container options for deploying on triton, throw an exception
            logger.info(f"No backend option provided for triton. Please use a valid backend type (vllm/djl) for the model.json file to be created. Backend provided={backend}")
            return
    except Exception as e:
        raise Exception(f"Error occurred while preparing files for the triton model container: {e}")


def _create_triton_service_neuron(model_id: str, 
                          num_model_copies: int, 
                          devices_per_model: int, 
                          image: str,
                          user: str, 
                          shm_size: str, 
                          env: List,
                          base_port: int, 
                          accelerator: constants.ACCELERATOR_TYPE, 
                          backend: constants.BACKEND) -> Tuple:
    """
    Creates the Triton service-specific part of the docker compose file. This function is responsible for handling the volume
    mounting, and other aspects to the docker compose file, such as the entrypoint command, port mapping, and more.
    """
    try:
        # initialize the cnames, services dictionary for triton on neuron, 
        # and more 
        cnames: List = []
        services: Dict = {}
        per_container_info_list: List = []
        ports_per_model_server: int = 3 # http, grps, metrics
        home = str(Path.home())
        triton_content_dir: Optional[str] = None
        triton_inference_script: Optional[str] = None
        dir_path_on_host: str = os.path.join(home, Path(model_id).name)

        if backend == constants.BACKEND.VLLM_BACKEND:
            logger.info(f"Setting triton_content_dir to {constants.TRITON_CONTENT_DIR_NAME_VLLM}")
            triton_content_dir = constants.TRITON_CONTENT_DIR_NAME_VLLM
            triton_inference_script = constants.TRITON_INFERENCE_SCRIPT_VLLM
        elif backend == constants.BACKEND.DJL_BACKEND:
            logger.info(f"Setting triton_content_dir to {constants.TRITON_CONTENT_DIR_NAME_DJL}")
            triton_content_dir = constants.TRITON_CONTENT_DIR_NAME_DJL
            triton_inference_script = constants.TRITON_INFERENCE_SCRIPT_DJL
        else: 
            logger.error(f"Triton content directory for {backend} not provided.")
            return
        
        # Iterate through the number of model copies and prepare the docker service for triton
        for i in range(num_model_copies):
            cname: str = f"fmbench_model_container_{i+1}"
            cnames.append(cname)

            device_offset = devices_per_model * i

            if accelerator == constants.ACCELERATOR_TYPE.NEURON:
                devices = [f"/dev/neuron{j + device_offset}:/dev/neuron{j}" for j in range(devices_per_model)]
                extra_env = []
            else:
                devices = None
                gpus = ",".join([str(j + device_offset) for j in range(devices_per_model)])
                extra_env = [f"NVIDIA_VISIBLE_DEVICES={gpus}"]
                env = env + extra_env

            # Setup Triton model content for each instance
            instance_dir: str = os.path.join(dir_path_on_host, f"i{i+1}")
            triton_instance_dir: str = os.path.join(instance_dir, triton_content_dir)
            os.makedirs(triton_instance_dir, exist_ok=True)
            current_dir: str = os.path.dirname(os.path.realpath(__file__))
            parent_dir: str = os.path.abspath(os.path.join(current_dir, os.pardir))
            triton_content: str = os.path.join(parent_dir, triton_content_dir)

            # Copy all files from triton_content to the instance's triton directory
            for item in os.listdir(triton_content):
                s = os.path.join(triton_content, item)
                d = os.path.join(triton_instance_dir, item)
                if os.path.isfile(s):
                    shutil.copy2(s, d)
                elif os.path.isdir(s):
                    shutil.copytree(s, d, dirs_exist_ok=True)
            # Set execute permissions for the script. The triton volumes contain the model repostory that
            # are mapped into the container and used during deployment
            # os.chmod((triton_instance_dir), 0o755)
            os.chmod(os.path.join(triton_instance_dir, os.path.basename(triton_inference_script)), 0o755)
            volumes = [f"{triton_instance_dir}:/scripts/triton:rw",
                       f"{triton_instance_dir}:/triton:rw"]
            # Create the Triton service
            total_neuron_cores = num_model_copies * devices_per_model * 2
            service = {
                cname: {
                    "image": image,
                    "container_name": cname,
                    "shm_size": shm_size,
                    "devices": devices,
                    "environment": env,
                    "volumes": volumes,
                    "ports": [f"{base_port + i*ports_per_model_server}:{base_port + i*ports_per_model_server}"],
                    "deploy": {"restart_policy": {"condition": "on-failure"}},
                    "command": f"{triton_inference_script} {model_id} {Path(model_id).name} {base_port + i*ports_per_model_server} {total_neuron_cores}"
                }
            }
            services.update(service)
            # for the triton container we could have multiple model servers within the same container
            nginx_server_lines = [f"        server {cname}:{base_port + i*ports_per_model_server};"]
            config_properties = CONFIG_PROPERTIES.format(port=(base_port + i*ports_per_model_server))
            per_container_info_list.append(dict(dir_path_on_host=instance_dir, 
                                                config_properties=config_properties, 
                                                container_name=cname,
                                                nginx_server_lines=nginx_server_lines))
    except Exception as e:
        logger.error(f"Error occurred while creating configuration files for triton: {e}")
        services, per_container_info_list = None, None
    nginx_command = "sh -c \"echo going to sleep for 240s && sleep 240 && echo after sleep && nginx -g \'daemon off;\' && echo started nginx\""
    return services, per_container_info_list, nginx_command


def _create_triton_service_gpu(model_id: str, 
                          num_model_copies: int, 
                          devices_per_model: int, 
                          image: str, 
                          user: str, 
                          shm_size: str, 
                          env: List,
                          base_port: int, 
                          accelerator: constants.ACCELERATOR_TYPE,
                          tp_degree: int,
                          batch_size: int) -> Tuple:
    """
    Creates the Triton service-specific part of the docker compose file. This function is responsible for handling the volume
    mounting, and other aspects to the docker compose file, such as the entrypoint command, port mapping, and more.
    """
    try:
        """
        the gpu version of services looks like this
        services:
    fmbench_model_container:
        image: nvcr.io/nvidia/tritonserver:24.08-trtllm-python-py3
        container_name: fmbench_model_container  # Name of the container
        runtime: nvidia  # Enables GPU support for the container
        shm_size: 12g  # Shared memory size
        ulimits:
          memlock: -1  # Remove memory locking limits
          stack: 67108864  # Set stack size
        ports:
        # sufficient for upto 4 instances of the model
        # add more if needed
        - 8000:8000
        - 8003:8003
        - 8006:8006
        - 8009:8009
        deploy:
          resources:
            reservations:
              devices:
                - driver: nvidia
                  count: all  # Use all available GPUs
                  capabilities: [gpu]
        volumes:
        - ${HOME}/tensorrtllm_backend:/tensorrtllm_backend
        - ${HOME}/${MODEL_ID}:/${MODEL_ID}
        - ${HOME}/engines:/engines
        - ${HOME}/deploy_on_triton/scripts:/scripts
        #network_mode: host  # Use the host's network stack
        tty: true  # Allocate a pseudo-TTY (interactive terminal)
        command: bash -c "/scripts/serve_model.sh ${MODEL_ID} ${TP_DEGREE} ${BATCH_SIZE} ${MODEL_COPIES} && bash"  # Run script and keep the container alive with bash
        restart: on-failure  # Ensure container restarts if it stops unexpectedly
        """
        cnames: List = []
        services: Dict = {}
        per_container_info_list: List = []
        home = str(Path.home())
        dir_path_on_host: str = os.path.join(home, Path(model_id).name)
        
        cname: str = FMBENCH_MODEL_CONTAINER_NAME
        ports_per_model_server: int = 3 # http, grps, metrics
        logger.info(f"_create_triton_service_gpu, model_id=\"{model_id}\", home=\"{home}\", dir_path_on_host=\"{dir_path_on_host}\", "
                    f"tp_degree=\"{tp_degree}\", batch_size=\"{batch_size}\", num_model_copies=\"{num_model_copies}\"")
        volumes = [f"{home}/tensorrtllm_backend:/tensorrtllm_backend",
                   f"{home}/{model_id}:/{model_id}",
                   f"{dir_path_on_host}/engines:/engines",
                   f"{dir_path_on_host}/scripts:/scripts"]
        
        # copy trtiton serving script
        triton_scripts_dir = os.path.join(dir_path_on_host, "scripts")
        os.makedirs(triton_scripts_dir, exist_ok=True)
        triton_serve_model_script_dst_path = os.path.join(triton_scripts_dir, constants.TRITON_SERVE_SCRIPT)
        script_dir_path = Path( __file__ ).parent.absolute()
        triton_serve_model_script_src_path = os.path.join(script_dir_path, constants.TRITON_SERVE_SCRIPT)
        logger.info(f"going to copy {triton_serve_model_script_src_path} to {triton_serve_model_script_dst_path}")
        shutil.copyfile(triton_serve_model_script_src_path, triton_serve_model_script_dst_path)
        st = os.stat(triton_serve_model_script_dst_path)
        os.chmod(triton_serve_model_script_dst_path, st.st_mode | stat.S_IEXEC)

        # setup the services section
        service = {
                cname: {
                    "image": image,
                    "container_name": cname,
                    "shm_size": shm_size,
                    "ulimits": {"memlock": -1, "stack": 67108864},
                    "volumes": volumes,
                    "ports": [f"{base_port + i*ports_per_model_server}:{base_port + i*ports_per_model_server}" for i in range(num_model_copies)],
                    "deploy": {"resources": {"reservations": {"devices": [{"driver": "nvidia", "count": "all", "capabilities": ['gpu']}]}}},
                    "tty": True,                    
                    "command": f"bash -c \"/scripts/{constants.TRITON_SERVE_SCRIPT} {model_id} {tp_degree} {batch_size} {num_model_copies} {base_port} && bash\"",
                    "restart": "on-failure"
                }
            }
        services.update(service)
        # for the triton container we could have multiple model servers within the same container
        nginx_server_lines = [f"        server {cname}:{base_port + i*ports_per_model_server};" for i in range(num_model_copies)]
        per_container_info_list.append(dict(dir_path_on_host=dir_path_on_host,
                                            config_properties=None,
                                            container_name=cname,
                                            nginx_server_lines=nginx_server_lines))
    except Exception as e:
        logger.error(f"Error occurred while creating configuration files for triton: {e}")
        services, per_container_info_list = None, None
    # ask the nginx lb to wait for a few minutes for the triton container to come up
    nginx_command = "sh -c \"echo going to sleep for 240s && sleep 240 && echo after sleep && nginx -g \'daemon off;\' && echo started nginx\""
    return services, per_container_info_list, nginx_command

def create_triton_service(model_id: str, 
                          num_model_copies: int, 
                          devices_per_model: int, 
                          image: str, 
                          user: str, 
                          shm_size: str, 
                          env: List,
                          base_port: int, 
                          accelerator: constants.ACCELERATOR_TYPE,
                          tp_degree: int,
                          batch_size: int,
                          backend: constants.BACKEND) -> Tuple:
    """
    Creates the Triton service-specific part of the docker compose file. This function is responsible for handling the volume
    mounting, and other aspects to the docker compose file, such as the entrypoint command, port mapping, and more.
    """

    if accelerator == constants.ACCELERATOR_TYPE.NEURON:
        logger.info(f"accelerator={accelerator}, calling the neuron version of this function")
        return _create_triton_service_neuron(model_id, 
                          num_model_copies, 
                          devices_per_model, 
                          image, 
                          user, 
                          shm_size, 
                          env,
                          base_port, 
                          accelerator,
                          backend) 
    else:
        logger.info(f"accelerator={accelerator}, calling the gpu version of this function")
        return _create_triton_service_gpu(model_id, 
                          num_model_copies, 
                          devices_per_model, 
                          image, 
                          user, 
                          shm_size, 
                          env,
                          base_port, 
                          accelerator,
                          tp_degree,
                          batch_size) 





import time
import json
import copy
import boto3
import logging
import sagemaker
import pandas as pd
from datetime import datetime
from typing import Dict, Optional
from fmbench.scripts import constants
from fmbench.utils import count_tokens
from sagemaker.predictor import Predictor
from sagemaker.serializers import JSONSerializer
from fmbench.scripts.stream_responses import get_response_stream
from fmbench.scripts.sagemaker_metrics import get_endpoint_metrics
from fmbench.scripts.fmbench_predictor import (FMBenchPredictor,
                                               FMBenchPredictionResponse)

# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Initialize the SageMaker runtime to support getting response streams
sagemaker_runtime = boto3.client('sagemaker-runtime')
sm_client = boto3.client("sagemaker")


class SageMakerPredictor(FMBenchPredictor):
    # overriding abstract method
    def __init__(self,
                 endpoint_name: str,
                 inference_spec: Optional[Dict],
                 metadata: Optional[Dict]):
        self._predictor: Optional[sagemaker.base_predictor.Predictor] = None
        self._endpoint_name: str = endpoint_name
        self._inference_spec: Dict = inference_spec
        self._variant_name: Optional[str] = None
        self._use_messages_api_format: Optional[bool] = None
        if metadata is not None:
            self._variant_name = metadata.get("variant_name")
            self._use_messages_api_format = metadata.get("use_messages_api_format")

        try:
            # Create a SageMaker Predictor object
            self._predictor = Predictor(
                endpoint_name=self._endpoint_name,
                sagemaker_session=sagemaker.Session(),
                serializer=JSONSerializer()
            )
        except Exception as e:
            logger.error(f"create_predictor, exception occured while creating predictor "
                         f"for endpoint_name={self._endpoint_name}, exception={e}")
        logger.info(f"__init__ _predictor={self._predictor}, "
                    f"_variant_name={self._variant_name}_"
                    f"inference_spec={self._inference_spec}, "
                    f"_use_messages_api_format={self._use_messages_api_format}")

    def get_prediction(self, payload: Dict) -> FMBenchPredictionResponse:
        response_json: Optional[Dict] = None
        response: Optional[str] = None
        latency: Optional[float] = None
        TTFT: Optional[float] = None
        TPOT: Optional[float] = None
        TTLT: Optional[float] = None
        prompt_tokens: Optional[int] = None
        completion_tokens: Optional[int] = None
        streaming: Optional[bool] = None
        stop_token: Optional[str] = None
        
        
        
        # represents the number of tokens in the prompt payload
        prompt_tokens = count_tokens(payload["inputs"])

        try:
            container_type = self._inference_spec.get("container_type")
            st = time.perf_counter()
            split_input_and_inference_params = None
            if self._inference_spec is not None:
                split_input_and_inference_params = self._inference_spec.get("split_input_and_parameters")
            response = None
            streaming = self._inference_spec.get("stream", False)
            if split_input_and_inference_params is True:
                response = self._predictor.predict(payload["inputs"],
                                                   self._inference_spec["parameters"])
            else:
                if self._use_messages_api_format is True:
                    # if needed in future add support for system prompt as well
                    # and multiple system/user prompts but all that is not needed for now
                    payload = {"messages": [{"role": "user",
                                             "content": payload["inputs"]}]}
                    # the final payload should look like this:
                    # {
                    #   "top_p": 0.9,
                    #   "max_tokens": 100,
                    #   "messages": [
                    #     {
                    #       "role": "user",
                    #       "content": "this is the prompt"
                    #     }
                    #   ]
                    # }
                    payload = payload | dict(self._inference_spec["parameters"])
                else:
                    # the final payload should look like this:
                    # {
                    #   "parameters": { 
                    #     "top_p": 0.9,
                    #     "max_tokens": 100
                    #    },
                    #   "inputs": "this is the prompt"
                    # }
                    if container_type == constants.CONTAINER_TYPE_HUGGINGFACE:
                        payload2 = copy.deepcopy(payload)
                        payload2['text_inputs'] = payload2.pop('inputs')
                        payload2['mode'] = "embedding"
                    else:
                        payload = payload | dict(parameters=self._inference_spec["parameters"])

            # if the response streaming is step, call the get_response stream on the 
            # sagemaker endpoint, else use the simple predict call
            if streaming is True:
                start_token = self._inference_spec.get("start_token", None)
                stop_token = self._inference_spec.get("stop_token", None)
                payload["stream"] = streaming
                logger.info(f"streaming={streaming}, calling invoke_endpoint_with_response_stream")
                response_stream = sagemaker_runtime.invoke_endpoint_with_response_stream(
                                                    EndpointName=self._endpoint_name,
                                                    Body=json.dumps(payload),
                                                    ContentType="application/json")
                response_dict = get_response_stream(response_stream['Body'],
                                                    st,
                                                    start_token,
                                                    stop_token,
                                                    is_sagemaker=True)
                TTFT = response_dict.get('TTFT')
                TPOT = response_dict.get('TPOT')
                TTLT = response_dict.get('TTLT')
                response = response_dict.get('response')
            else:
                logger.info(f"streaming={streaming}, calling predict")
                if container_type:
                    response = self._predictor.predict(payload2)
                    logger.info("Running predictor with HuggingFace Container for Embedding model")
                else:
                    response = self._predictor.predict(payload)
                    logger.info("Running predictor for Foundation model")

            latency = time.perf_counter() - st
            if isinstance(response, bytes):
                response = response.decode('utf-8')
            response_json = json.loads(response)

            # we want to set the "generated_text" key in the response
            if isinstance(response_json, list):
                response_json = response_json[0]
                # add a key called completion, if not there
                if response_json.get("generated_text") is None:
                    # look for predicted label and set that as generated text
                    if response_json.get("predicted_label") is not None:
                        response_json["generated_text"] = response_json.get("predicted_label")
                    else:
                        logger.error("response_json is list but did not contain generated_text or predicted_label, dont know how to handle this")
            elif isinstance(response_json, dict):
                choices = response_json.get("choices")
                if choices is not None:
                    if isinstance(choices, list):
                        response_json = response_json["choices"][0]["message"]
                        if response_json.get("generated_text") is None:
                            if response_json.get("content") is not None:
                                response_json["generated_text"] = response_json.get("content")
                            else:
                                logger.error(f"response_json is a dict, choices is a list, but response_json does not contain generated_text, dont know how to handle this")
                        else:
                            logger.error(f"response_json is a dict, choices is a list, but generated_text ia None, dont know how to handle this")
                    else:
                        logger.error(f"response_json is a dict, but choices is not a list but rather it is {type(choices)}, dont know how to handle this")
                else:
                    if container_type == constants.CONTAINER_TYPE_HUGGINGFACE:
                        response_json["generated_text"] = response_json.get("embedding")
                        completion_tokens = len(response_json.get("generated_text"))
                    # logger.error(f"response_json is a dict, but does not contain choices, dont know how to handle this")
            else:
                logger.error(f"response_json data type is {type(response_json)}, dont know how to handle this")
            # counts the completion tokens for the model using the default/user provided tokenizer
            if not completion_tokens:
                completion_tokens = count_tokens(response_json.get("generated_text"))

        except Exception as e:
            logger.error(f"get_prediction, exception occurred while getting prediction for payload={payload} "
                         f"from predictor={self._endpoint_name}, response={response}, exception={e}")
        return FMBenchPredictionResponse(response_json=response_json,
                                         latency=latency,
                                         time_to_first_token=TTFT,
                                         time_per_output_token=TPOT,
                                         time_to_last_token=TTLT,
                                         completion_tokens=completion_tokens,
                                         prompt_tokens=prompt_tokens)

    @property
    def endpoint_name(self) -> str:
        """The endpoint name property."""
        return self._endpoint_name

    def calculate_cost(self,
                       instance_type: str,
                       instance_count: int,
                       pricing: Dict,
                       duration: float,
                       prompt_tokens: int,
                       completion_tokens: int) -> float:
        """Calculate the cost of each experiment run."""
        experiment_cost: Optional[float] = None
        try:
            instance_based_pricing = pricing['pricing']['instance_based']
            hourly_rate = instance_based_pricing.get(instance_type, None)
            logger.info(f"the hourly rate for running on {instance_type} is {hourly_rate}, "
                        f"instance_count={instance_count}")
            # calculating the experiment cost for instance based pricing
            instance_count = instance_count if instance_count else 1
            experiment_cost = (hourly_rate / 3600) * duration * instance_count
        except Exception as e:
            logger.error(f"exception occurred during experiment cost calculation, exception={e}")
        return experiment_cost

    def get_metrics(self,
                    start_time: datetime,
                    end_time: datetime,
                    period: int = 60) -> pd.DataFrame:
        return get_endpoint_metrics(self._endpoint_name, self._variant_name, start_time, end_time)
        
    def shutdown(self) -> None:
        """Represents the function to shutdown the predictor
           cleanup the endpooint/container/other resources
        """
        try:
            ep_name = self.endpoint_name
            ## Describe the model endpoint 
            logger.info(f"Going to describe the endpoint -> {ep_name}")
            resp = sm_client.describe_endpoint(EndpointName=ep_name)

            ## If the given model endpoint is in service, delete it 
            if resp['EndpointStatus'] == 'InService':
                logger.info(f"going to delete {ep_name}")
                ## deleting the model endpoint
                sm_client.delete_endpoint(EndpointName=ep_name)
                logger.info(f"deleted {ep_name}")
                return True
        except Exception as e:
            logger.error(f"error deleting endpoint={ep_name}, exception={e}")
            return False
    
    @property
    def inference_parameters(self) -> Dict:
        """The inference parameters property."""
        return self._inference_spec.get("parameters")


    @property
    def platform_type(self) -> Dict:
        """The inference parameters property."""
        return constants.PLATFORM_SAGEMAKER
    
def create_predictor(endpoint_name: str, inference_spec: Optional[Dict], metadata: Optional[Dict]):
    return SageMakerPredictor(endpoint_name, inference_spec, metadata)



"""
DJL specific code
"""
import os
from pathlib import Path
import fmbench.scripts.constants as constants
from typing import Dict, List, Optional, Tuple, Union
from fmbench.scripts.inference_containers.utils import (STOP_AND_RM_CONTAINER,
                                                        FMBENCH_MODEL_CONTAINER_NAME)

CONFIG_PROPERTIES: str = """
            inference_address=http://0.0.0.0:{port}
            management_address=http://0.0.0.0:{port}
            cluster_address=http://0.0.0.0:8888
            model_store=/opt/ml/model
            load_models=ALL
            """

def create_script(region, image_uri, model_id, model_name, env_str, privileged_str, hf_token, directory):
    """
    Script for running the docker container for the inference server
    """
    script = f"""#!/bin/sh
        echo "Going to download model container"
        echo "Content in docker command: {region}, {image_uri}, {model_name}"

        # Login to AWS ECR and pull the Docker image
        aws ecr get-login-password --region {region} | docker login --username AWS --password-stdin {image_uri}
        docker pull {image_uri}       

        # Run the new Docker container with specified settings
        cd {directory}
        # shutdown existing docker compose
        docker compose down
        {STOP_AND_RM_CONTAINER}
        docker compose up -d
        cd -
        echo "started docker compose in daemon mode"
    """
    return script

def create_djl_service(model_id: str, 
                        num_model_copies: int, 
                        devices_per_model: int, 
                        image: str, 
                        user: str, 
                        shm_size: str, 
                        env: List, 
                        base_port: int, 
                        accelerator: constants.ACCELERATOR_TYPE) -> Tuple:
    """
    Creates the service for DJL, setting up devices and other configurations.
    """
    try:
        cnames: List = []
        services: Dict = {}
        per_container_info_list: List = []
        home = str(Path.home())
        dir_path_on_host: str = os.path.join(home, Path(model_id).name)
        
        # Iterate through the number of model copies and prepare the docker service for djl
        for i in range(num_model_copies):
            cname = f"fmbench_model_container_{i+1}"
            cnames.append(cname)

            device_offset = devices_per_model * i

            if accelerator == constants.ACCELERATOR_TYPE.NEURON:
                devices = [f"/dev/neuron{j + device_offset}:/dev/neuron{j}" for j in range(devices_per_model)]
                extra_env = []
            else:
                devices = None
                gpus = ",".join([str(j + device_offset) for j in range(devices_per_model)])
                extra_env = [f"NVIDIA_VISIBLE_DEVICES={gpus}"]
                env = env + extra_env

            volumes = [f"{dir_path_on_host}/i{i+1}:/opt/ml/model:ro",
                       f"{dir_path_on_host}/i{i+1}/conf:/opt/djl/conf:ro",
                       f"{dir_path_on_host}/i{i+1}/model_server_logs:/opt/djl/logs"]
            # compute the port
            port = base_port + i

            service = {
                cname: {
                    "image": image,
                    "container_name": cname,
                    "user": user,
                    "shm_size": shm_size,
                    "devices": devices,
                    "environment": env,
                    "volumes": volumes,
                    "ports": [f"{port}:{port}"],
                    "deploy": {"restart_policy": {"condition": "on-failure"}}
                }
            }
            if accelerator == constants.ACCELERATOR_TYPE.NVIDIA:
                service[cname]['runtime'] = constants.ACCELERATOR_TYPE.NVIDIA.value
                service[cname].pop("devices")
            services.update(service)
            config_properties = CONFIG_PROPERTIES.format(port=port)


            nginx_server_lines = [f"        server {cname}:{port};"]
    
            per_container_info_list.append(dict(dir_path_on_host=f"{dir_path_on_host}/i{i+1}",
                                                config_properties=config_properties,
                                                container_name=cname,
                                                nginx_server_lines=nginx_server_lines))
    except Exception as e:
        logger.error(f"Error occurred while generating configuration files for djl/vllm: {e}")
        services, per_container_info_list=None, None
    # since the djl container comes up immediately and the nginx lb does not have to wait
    # so the nginx command is a noop
    nginx_command = None
    return services, per_container_info_list, nginx_command



from enum import Enum
from typing import List

CONTAINER_TYPE_DJL: str = 'djl'
CONTAINER_TYPE_VLLM: str = 'vllm'
CONTAINER_TYPE_TRITON: str = 'triton'
CONTAINER_TYPE_OLLAMA: str = 'ollama'
CONTAINER_TYPE_HUGGINGFACE: str = 'huggingface'
TRITON_INFERENCE_SCRIPT_VLLM: str = '/scripts/triton/triton-vllm-neuronx.sh'
TRITON_INFERENCE_SCRIPT_DJL: str = '/scripts/triton/triton-djl-python-neuronx.sh'
TRITON_CONTENT_DIR_NAME_VLLM: str = 'triton/vllm'
TRITON_CONTENT_DIR_NAME_DJL: str = 'triton/djl'
TRITON_SERVE_SCRIPT: str = "triton_serve_model.sh"
AWS_CHIPS_PREFIX_LIST: List[str] = ["inf2", "trn1"]
IS_NEURON_INSTANCE = lambda instance_type: any([instance_type.startswith(p) for p in AWS_CHIPS_PREFIX_LIST])

class ACCELERATOR_TYPE(str, Enum):
    NEURON = 'neuron'
    NVIDIA = "nvidia"

class BACKEND(str, Enum):
    VLLM_BACKEND = 'vllm'
    DJL_BACKEND = 'djl'
    TENSORRT_BACKEND = 'tensorrt'

class MODEL_COPIES(str, Enum):
    AUTO = 'auto'
    MAX = "max"
    
    
# These variables represent the platform where a specific
# endpoint is deployed.
PLATFORM_SAGEMAKER: str = "sagemaker"
PLATFORM_EKS: str = "eks"
PLATFORM_EC2: str = "ec2"
PLATFORM_BEDROCK: str = "bedrock"
PLATFORM_EXTERNAL: str = "external"

# inference server listen port
BASE_PORT_FOR_CONTAINERS: int = 8000
LISTEN_PORT: int = 8080

# This is the file where the EC2 instance utilization metrics are stored
EC2_SYSTEM_METRICS_FNAME: str = "EC2_system_metrics.csv"
EC2_UTILIZATION_METRICS_INTERVAL: int = 5



"""
Utility functions common across inference containers
"""
import json
import logging
import subprocess
import fmbench.scripts.constants as constants
from typing import Dict, List, Optional, Tuple, Union

logging.basicConfig(format='[%(asctime)s] p%(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

FMBENCH_MODEL_CONTAINER_NAME: str = "fmbench_model_container"

STOP_AND_RM_CONTAINER = f"""
    # Attempt to stop and remove the container up to 3 times if container exists
        if [ -n "$(docker ps -aq --filter "name={FMBENCH_MODEL_CONTAINER_NAME}")" ]; then
            for i in {{1..3}}; do
                echo "Attempt $i to stop and remove the container: {FMBENCH_MODEL_CONTAINER_NAME}"
                
                # Stop the container
                docker ps -q --filter "name={FMBENCH_MODEL_CONTAINER_NAME}" | xargs -r docker stop
                
                # Wait for 5 seconds
                sleep 5
                
                # Remove the container
                docker ps -aq --filter "name={FMBENCH_MODEL_CONTAINER_NAME}" | xargs -r docker rm
                
                # Wait for 5 seconds
                sleep 5
                
                # Check if the container is removed
                if [ -z "$(docker ps -aq --filter "name={FMBENCH_MODEL_CONTAINER_NAME}")" ]; then
                    echo "Container {FMBENCH_MODEL_CONTAINER_NAME} successfully stopped and removed."
                    break
                else
                    echo "Container {FMBENCH_MODEL_CONTAINER_NAME} still exists, retrying..."
                fi
            done
        else
            echo "Container {FMBENCH_MODEL_CONTAINER_NAME} does not exist. No action taken."
        fi
    """


def get_accelerator_type() -> constants.ACCELERATOR_TYPE:
    # first check if this is an NVIDIA instance or a AWS Chips instance
    # we do this by checking for nvidia-smi and neuron-ls, they should be
    # present on NVIDIA and AWS Chips instances respectively
    NVIDIA_SMI: str = "nvidia-smi"
    NEURON_LS: str = "neuron-ls"
    # if the utility is not present the return from the whereis command is like
    # 'neuron-ls:\n' or 'nvidia-smi:\n' otherwise it is like 
    # 'nvidia-smi: /usr/bin/nvidia-smi /usr/share/man/man1/nvidia-smi.1.gz\n'
    utility_present = lambda x: subprocess.check_output(["whereis", x]).decode() != f"{x}:\n"
    is_nvidia = utility_present(NVIDIA_SMI)
    is_neuron = utility_present(NEURON_LS)
    logger.info(f"is_nvidia={is_nvidia}, is_neuron={is_neuron}")
    if is_nvidia is True:
        return constants.ACCELERATOR_TYPE.NVIDIA
    else:
        return constants.ACCELERATOR_TYPE.NEURON


def get_model_copies_to_start_neuron(tp_degree: int, model_copies: str) -> Tuple[int, int]:
    logger.info(f"get_model_copies_to_start_neuron, tp_degree={tp_degree}, model_copies={model_copies}")
    # get the number of neuron cores
    cmd = ["neuron-ls -j"]
    process = subprocess.Popen(cmd,
                               stdout=subprocess.PIPE, 
                               stderr=subprocess.PIPE,
                               text=True,
                               shell=True)
    std_out, std_err = process.communicate()
    logger.info(std_out.strip())
    logger.info(std_err)

    if std_err != '':
        logger.error("error determining neuron info, exiting")
        return None

    # convert output to a dict for easy parsing
    neuron_info = json.loads(std_out)
    num_neuron_devices = len(neuron_info)
    num_neuron_cores = 2 * num_neuron_devices

    if model_copies == constants.MODEL_COPIES.AUTO:
        model_copies_as_int = 1
        devices_per_model = num_neuron_devices
        logger.info(f"model copies is set to {model_copies}, model_copies_as_int={model_copies_as_int}, "
                    f"devices_per_model={devices_per_model}")
        return model_copies_as_int, devices_per_model
              
    # tensor parallelism requires as many neuron cores as tp degree
    # and the number of devices required is half of number of cores
    neuron_devices_needed_per_model_copy = int(tp_degree / 2)
    model_copies_possible = int(num_neuron_devices/neuron_devices_needed_per_model_copy)
    # if model_copies is max then load as many copies as possible
    if isinstance(model_copies, str):
        if model_copies == constants.MODEL_COPIES.MAX:
            model_copies = model_copies_possible
            logger.info(f"model_copies was set to \"{constants.MODEL_COPIES.MAX}\", "
                        f"this instance can support a max of {model_copies} model copies, "
                        f"going to load {model_copies} copies")
        else:            
            logger.info(f"model_copies set to a str value={model_copies}, "
                        f"will see if we can load these many models")
            model_copies = int(model_copies)
    else:
        logger.info(f"model_copies set to a numerical value={model_copies}, "
                    f"will see if we can load these many models")

    num_devices_needed = model_copies * neuron_devices_needed_per_model_copy
    
    logger.info(f"model_copies={model_copies}, tp_degree={tp_degree},\n"
                f"num_neuron_devices={num_neuron_devices}, num_neuron_cores={num_neuron_cores},\n"
                f"neuron_devices_needed_per_model_copy={neuron_devices_needed_per_model_copy}, num_devices_needed={num_devices_needed},\n"
                f"model_copies_possible={model_copies_possible}")

    if model_copies_possible < model_copies:
        logger.error(f"model_copies={model_copies} but model_copies_possible={model_copies_possible}, "
                     f"setting model_copies to max possible for this instance which is {model_copies_possible}")
        model_copies = model_copies_possible
    else:
        logger.error(f"model_copies={model_copies} and model_copies_possible={model_copies_possible}, "
                     f"it is possible to run {model_copies} models, going with that")

    return model_copies, neuron_devices_needed_per_model_copy

def get_model_copies_to_start_nvidia(tp_degree: int, model_copies: str) -> Tuple[int, int]:
    logger.info(f"get_model_copies_to_start_nvidia, tp_degree={tp_degree}, model_copies={model_copies}")
    # get the number of neuron cores
    cmd = ["nvidia-smi --query-gpu=name --format=csv,noheader | wc -l"]
    process = subprocess.Popen(cmd,
                               stdout=subprocess.PIPE, 
                               stderr=subprocess.PIPE,
                               text=True,
                               shell=True)
    std_out, std_err = process.communicate()
    logger.info(std_out.strip())
    logger.info(std_err)

    if std_err != '':
        logger.error("error determining neuron info, exiting")
        return None

    num_gpus = int(std_out.strip())

    if model_copies == constants.MODEL_COPIES.AUTO:
        model_copies_as_int = 1
        devices_per_model = num_gpus
        logger.info(f"model copies is set to {model_copies}, model_copies_as_int={model_copies_as_int}, "
                    f"devices_per_model={devices_per_model}")
        return model_copies_as_int, devices_per_model
    
    # tensor parallelism requires as many gpus as tp degree
    gpus_needed_per_model_copy = tp_degree
    model_copies_possible = int(num_gpus/gpus_needed_per_model_copy)
    # if model_copies is max then load as many copies as possible
    if isinstance(model_copies, str):
        if model_copies == constants.MODEL_COPIES.MAX:
            model_copies = model_copies_possible
            logger.info(f"model_copies was set to \"{constants.MODEL_COPIES.MAX}\", "
                        f"this instance can support a max of {model_copies} model copies, "
                        f"going to load {model_copies} copies")
        else:            
            logger.info(f"model_copies set to a str value={model_copies}, "
                        f"will see if we can load these many models")
            model_copies = int(model_copies)
    else:
        logger.info(f"model_copies set to a numerical value={model_copies}, "
                    f"will see if we can load these many models")

    num_gpus_needed = model_copies * gpus_needed_per_model_copy
    
    logger.info(f"model_copies={model_copies}, tp_degree={tp_degree},\n"
                f"num_gpus={num_gpus},\n"
                f"gpus_needed_per_model_copy={gpus_needed_per_model_copy}, num_gpus_needed={num_gpus_needed},\n"
                f"model_copies_possible={model_copies_possible}")

    if model_copies_possible < model_copies:
        logger.error(f"model_copies={model_copies} but model_copies_possible={model_copies_possible}, "
                     f"setting model_copies to max possible for this instance which is {model_copies_possible}")
        model_copies = model_copies_possible
    else:
        logger.info(f"model_copies={model_copies} and model_copies_possible={model_copies_possible}, "
                     f"it is possible to run {model_copies} models, going with that")

    return model_copies, gpus_needed_per_model_copy




import pandas as pd
from datetime import datetime
from typing import Dict, Optional
from abc import ABC, abstractmethod, abstractproperty


class FMBenchPredictor(ABC):

    @abstractmethod
    def __init__(self,
                 endpoint_name: str,
                 inference_spec: Optional[Dict],
                 metadata: Optional[Dict]):
        pass

    @abstractmethod
    def get_prediction(self, payload: Dict) -> Dict:
        pass

    @abstractmethod
    def calculate_cost(self,
                       instance_type: str,
                       instance_count: int,
                       config: Dict,
                       duration: float,
                       metrics: Dict) -> float:
        """Represents the function to calculate the
           cost of each experiment run.
        """
        pass
    
    @abstractmethod
    def get_metrics(self,
                    start_time: datetime,
                    end_time: datetime,
                    period: int = 60) -> pd.DataFrame:
        """Represents the function to calculate the
           metrics for each endpoint
        """
        pass

    @abstractmethod
    def shutdown(self) -> None:
        """Represents the function to shutdown the predictor
           cleanup the endpooint/container/other resources
        """
        return None

    @abstractproperty
    def endpoint_name(self) -> str:
        """The endpoint name property."""
        pass

    @abstractproperty
    def inference_parameters(self) -> Dict:
        """The inference parameters property."""
        pass

    @abstractproperty
    def platform_type(self) -> Dict:
        """The inference parameters property."""
        pass


class FMBenchPredictionResponse(dict):

    def __init__(self, *k, **kwargs):
        self.__dict__ = self
        self.__dict__['response_json'] = kwargs['response_json']
        self.__dict__['latency'] = kwargs['latency']
        self.__dict__['prompt_tokens'] = kwargs['prompt_tokens']
        self.__dict__['completion_tokens'] = kwargs['completion_tokens']
        super().__init__(*k, **kwargs)



# script that prepares a directory structure to run multiple copies of a model
# using the DJL container. This code is written for Neuron for now but would be modified
# to run on GPUs as well and then maybe other inference containers as well.
# It creates a docker_compose.yml file that creates multiple containers and a load balancer.
# which provides a single endpoint to external applications that want to use these containers.

import os
import json
import yaml
import shutil
import docker
import logging
import subprocess
from os import listdir
from pathlib import Path
from os.path import isfile, join
import fmbench.scripts.constants as constants
from typing import Dict, List, Optional, Tuple, Union
import fmbench.scripts.inference_containers.djl as djl
import fmbench.scripts.inference_containers.vllm as vllm
import fmbench.scripts.inference_containers.triton as triton
import fmbench.scripts.inference_containers.utils as ic_utils

logging.basicConfig(format='[%(asctime)s] p%(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

def _create_config_files(model_id: str, 
                        num_model_copies: int, 
                        devices_per_model: int, 
                        image: str, 
                        tp_degree: int,
                        batch_size: int,
                        user: str, 
                        shm_size: str, 
                        env: List, 
                        accelerator: constants.ACCELERATOR_TYPE,
                        dir_path: str,
                        listen_port: int, 
                        backend: constants.BACKEND) -> Tuple:
    """
    Creates the docker compose yml file, nginx config file, these are common to all containers
    and then creates individual config.properties files

    version: '3.8'

    services:
      fmbench_model_container_1:
        image: 763104351884.dkr.ecr.us-east-1.amazonaws.com/djl-inference:0.29.0-neuronx-sdk2.19.1
        # deepjavalibrary/djl-serving:0.29.0-pytorch-inf2 #
        container_name: fmbench_model_container_1
        user: djl
        shm_size: 12GB
        devices:
        - "/dev/neuron0:/dev/neuron0"
        - "/dev/neuron1:/dev/neuron1"
        - "/dev/neuron2:/dev/neuron2"
        - "/dev/neuron3:/dev/neuron3"
        - "/dev/neuron4:/dev/neuron4"
        - "/dev/neuron5:/dev/neuron5"
        environment:
        - MODEL_LOADING_TIMEOUT=2400
        - HF_TOKEN=<hf_token>
        volumes:
        - /home/ubuntu/Mistral-7B-Instruct-v0.2-i1:/opt/ml/model:ro
        - /home/ubuntu/Mistral-7B-Instruct-v0.2-i1/conf:/opt/djl/conf:ro
        - /home/ubuntu/Mistral-7B-Instruct-v0.2-i1/model_server_logs:/opt/djl/logs
        ports:
        - "8081:8081"
        deploy:
        restart_policy:
            condition: on-failure

      fmbench_model_container_2:
        image: 763104351884.dkr.ecr.us-east-1.amazonaws.com/djl-inference:0.29.0-neuronx-sdk2.19.1    
        container_name: fmbench_model_container_2
        user: djl
        shm_size: 12GB
        devices:
        - "/dev/neuron6:/dev/neuron0"
        - "/dev/neuron7:/dev/neuron1"
        - "/dev/neuron8:/dev/neuron2"
        - "/dev/neuron9:/dev/neuron3"
        - "/dev/neuron10:/dev/neuron4"
        - "/dev/neuron11:/dev/neuron5"
        environment:
        - MODEL_LOADING_TIMEOUT=2400
        - HF_TOKEN=<hf_token>
        volumes:
        - /home/ubuntu/Mistral-7B-Instruct-v0.2-i2:/opt/ml/model:ro
        - /home/ubuntu/Mistral-7B-Instruct-v0.2-i2/conf:/opt/djl/conf:ro
        - /home/ubuntu/Mistral-7B-Instruct-v0.2-i2/model_server_logs:/opt/djl/logs
        ports:
        - "8082:8082"
        deploy:
        restart_policy:
            condition: on-failure

      loadbalancer:
        image: nginx:alpine
        container_name: fmbench_model_container_loadbalancer
        ports:
        - "8080:80"
        volumes:
        - ./nginx.conf:/etc/nginx/nginx.conf:ro
        depends_on:
        - fmbench_model_container_1
        - fmbench_model_container_2
        deploy:
        placement:
            constraints: [node.role == manager]
    """

    # do we need to create a load balancer?
    # if the model copies or the model containers are > 1
    # then yes..note that we could have a situation where multiple
    # model servers are created within a single container so go with
    # model copies as the guide rather than containers
    lb_needed = True if num_model_copies > 1 else False
    try:
        if lb_needed is False:
            base_port = listen_port
            logger.info(f"lb_needed={lb_needed}, going to make the container listen on {listen_port} instead of "
                        f"{base_port}")
        else:
            base_port = constants.BASE_PORT_FOR_CONTAINERS
        if user == constants.CONTAINER_TYPE_TRITON:            
            logger.info(f"container type is {constants.CONTAINER_TYPE_TRITON}, accelerator={accelerator}. "
                        f"preparing the service for docker-compose")
            services, per_container_info_list, nginx_command = triton.create_triton_service(model_id, 
                                                                    num_model_copies, 
                                                                    devices_per_model, 
                                                                    image, 
                                                                    user, 
                                                                    shm_size,
                                                                    env, 
                                                                    base_port,
                                                                    accelerator,
                                                                    tp_degree,
                                                                    batch_size, 
                                                                    backend)
        else:
            logger.info(f"Container type is {user}. Preparing the service for docker-compose")
            services, per_container_info_list, nginx_command = djl.create_djl_service(model_id, 
                                                                    num_model_copies, 
                                                                    devices_per_model, 
                                                                    image, 
                                                                    user, 
                                                                    shm_size, 
                                                                    env, 
                                                                    base_port,
                                                                    accelerator)
        # Load balancer setup
        if lb_needed is True:
            fmbench_container_names = [ci['container_name'] for ci in per_container_info_list]
            # if this is a triton container and 
            lb = {
                "image": "nginx:alpine",
                "container_name": "fmbench_model_container_load_balancer",
                "command": nginx_command,
                "restart": "on-failure",
                "ports": [f"{listen_port}:80"],
                "volumes": ["./nginx.conf:/etc/nginx/nginx.conf:ro"],
                "depends_on": fmbench_container_names,
                "deploy": {"placement": {"constraints": ['node.role == manager']}}
            }
            if nginx_command is None:
                lb.pop("command")
            services.update({"loadbalancer": lb})
        else:
            logger.info(f"not creating load balancer")
        docker_compose = {"services": services}

        # nginx.conf generation
        if lb_needed is True:
            # the container info list contains the nginx.conf "server fmbench_model_Container_1:8000" type lines
            # as well ..but..not all inference containers are the same..for DJL we would have one model server
            # per container (or even if we have multiple they listen on the same port (?) so we dont have to care
            # about it in the nginx ln) but in Triton the model servers listen on different ports and so even if
            # there is a single triton container it could have say 4 model servers on different ports..all this 
            # to say that the nginx_conf_lines is a list and so we have a list of list situation happening here and 
            # we need to flatten this list of list
            
            nginx_server_lines = [l for ci in per_container_info_list for l in ci['nginx_server_lines']]
            # make it into a string
            nginx_server_lines = "\n".join(nginx_server_lines)

            nginx_config = f"""
### Nginx Load Balancer
events {{}}
http {{
    upstream fmcluster {{
{nginx_server_lines}
    }}
    server {{
        listen 80;
        location / {{
            proxy_pass http://fmcluster;
        }}
    }}
}}
"""
        else:
            nginx_config = None
    except Exception as e:
        logger.error(f"Error occurred while generating config files: {e}")
        docker_compose, nginx_config, per_container_info_list = None, None, None
    return docker_compose, nginx_config, per_container_info_list


def prepare_docker_compose_yml(model_id: str,
                               model_copies: str,
                               inference_params: Dict,
                               image: str,
                               user: str,
                               shm_size: str,
                               env: Dict,
                               serving_properties: Optional[str],
                               dir_path: str,
                               listen_port: int) -> int:

    # convert the env dict to a list of k=v pairs
    env_as_list = []
    if env is not None:
        for k,v in env.items():
            env_as_list.append(f"{k}={v}")

    # this is specific to the triton container
    # then default it to None and let the inference
    # container the best default values for parameters
    container_params: Optional[Dict] = inference_params.get('container_params', dict())
    logger.info(f"container_params that will be used for {image} image: {container_params}")  

    # the following code handles backwards compatibility where batch_Size and tp_degree are now specified
    # as part of container params but if not found then it falls back to the original scheme where
    # they were specified as part of the inference params
    batch_size: int = container_params.get('batch_size', inference_params.get('batch_size', 4))
    tp_degree: int = container_params.get('tp_degree', inference_params.get('tp_degree', 1))

    backend: Optional[str] = inference_params.get('backend')
    triton_content: Optional[str] = None
    if backend is not None:
        if backend == constants.BACKEND.VLLM_BACKEND:
            triton_content = constants.TRITON_CONTENT_DIR_NAME_VLLM
        elif backend == constants.BACKEND.DJL_BACKEND:
            triton_content = constants.TRITON_CONTENT_DIR_NAME_DJL
        else:
            logger.error(f"No backend={backend} provided")
    else:
        logger.error(f"No backend={backend} provided")

    # first check if this is an NVIDIA instance or a AWS Chips instance    
    accelerator = ic_utils.get_accelerator_type()

    if accelerator == constants.ACCELERATOR_TYPE.NVIDIA:
        model_copies_as_int, devices_per_model = ic_utils.get_model_copies_to_start_nvidia(tp_degree, model_copies)
    else:
        model_copies_as_int, devices_per_model = ic_utils.get_model_copies_to_start_neuron(tp_degree, model_copies)
    
    # special handling for Triton && Neuron
    if user == constants.CONTAINER_TYPE_TRITON and accelerator == constants.ACCELERATOR_TYPE.NEURON:
        logger.info(f"running Triton container on neuron, creating specific config files for this combination")
        current_dir: str = os.path.dirname(os.path.realpath(__file__))
        triton_content: str = os.path.join(current_dir, triton_content)
        # handles custom tensor pd, batch size into the model repository files
        triton.handle_triton_serving_properties_and_inf_params(triton_content, 
                                                               tp_degree, 
                                                               container_params,
                                                               model_id,
                                                               backend)

    # create the docker compose and nginx.conf file in the top level
    # directory path for this model
    dir_path = os.path.join(dir_path, Path(model_id).name)
    os.makedirs(dir_path, exist_ok=True)
    docker_compose, nginx_config, per_container_info_list = _create_config_files(model_id,
                                                                                 model_copies_as_int,
                                                                                 devices_per_model,
                                                                                 image,
                                                                                 tp_degree,
                                                                                 batch_size,
                                                                                 user,
                                                                                 shm_size,
                                                                                 env_as_list,
                                                                                 accelerator,
                                                                                 dir_path,
                                                                                 listen_port, 
                                                                                 backend)

    yaml.Dumper.ignore_aliases = lambda self, data: True
    docker_compose_yaml = yaml.dump(docker_compose)


    # ready to create the common files: docker_compose.yml and nginx.conf
    # any inference server specific files are created by the _create_config_files function
    # which in turn calls inference server specific functions
    dc_path = os.path.join(dir_path, "docker-compose.yml")
    logger.info(f"writing docker compose to {dc_path}, contents --->\n{docker_compose_yaml}")    
    Path(dc_path).write_text(docker_compose_yaml)

    # create nginx.conf file
    if nginx_config is not None:
        ngc_path = os.path.join(dir_path, "nginx.conf")
        logger.info(f"writing nginx conf to {ngc_path}, contents --->\n{nginx_config}")    
        Path(ngc_path).write_text(nginx_config)
    else:
        logger.info(f"model_copies={model_copies}, nginx_config={nginx_config}, "
                    f"not creating nginx.conf")

    # create sub directories for each model instance
    for idx, pc in enumerate(per_container_info_list):
        logger.info(f"creating files for container {idx+1} of {len(per_container_info_list)}...")
        # create serving.properties
        if serving_properties is not None:
            os.makedirs(pc["dir_path_on_host"], exist_ok=True)
            sp_fpath = os.path.join(pc["dir_path_on_host"], "serving.properties")
            logger.info(f"writing {serving_properties} to {sp_fpath}")
            Path(sp_fpath).write_text(serving_properties)
        else:
            logger.info(f"serving_properties={serving_properties}, skipping")

        # write config.properties, only applicable to DJL
        if pc['config_properties'] is not None:
            conf_dir = os.path.join(pc["dir_path_on_host"], "conf")
            os.makedirs(conf_dir, exist_ok=True)
            cp_fpath = os.path.join(conf_dir, "config.properties")
            logger.info(f"writing {pc['config_properties']} to {cp_fpath}")
            Path(cp_fpath).write_text(pc['config_properties'])
        else:
            logger.info(f"config_properties={pc['config_properties']}, skipping")

    return model_copies_as_int



DEFAULT_BUCKET_WRITE: str = 'sagemaker-fmbench-write'
DEFAULT_BUCKET_READ: str = 'sagemaker-fmbench-read'
DEFAULT_LOCAL_WRITE: str = 'fmbench-write'
DEFAULT_LOCAL_READ: str = 'fmbench-read'
MAX_REQ_CNT: int = 105
MIN_REQ_CNT: int = 5



import os
import torch
import logging
import argparse
from transformers.models.opt import OPTForCausalLM
from transformers_neuronx.module import save_pretrained_split
from transformers import AutoConfig, AutoModelForCausalLM, AutoTokenizer

MODEL_REPO: str = "meta-llama"
MODEL_ID: str = "Meta-Llama-3-1-70bB-Instruct"
NEURON_VER: str = "2.18.2"

root = logging.getLogger()
if root.handlers:
    for handler in root.handlers:
        root.removeHandler(handler)
logging.basicConfig(format='[%(asctime)s] p%(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

if __name__ == "__main__":
    
    if 'HF_TOKEN' not in os.environ:
        logger.info('Hugging Face Hub token is missing')
        exit(-1)

    # Define and parse command-line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-name", "-m", 
        type=str, 
        default=f"{MODEL_REPO}/{MODEL_ID}",
        help="HuggingFace model name"
    )
    parser.add_argument(
        "--save-path", "-s",
        type=str,
        default=f"../{NEURON_VER}/model_store/{MODEL_ID}/{MODEL_ID}-split/",
        help="Output directory for downloaded model files",
    )
    args = parser.parse_args()
    logger.info(f"args={args}")

    save_path = os.makedirs(args.save_path, exist_ok=True)
    logger.info(f"Save path defined for the model: {save_path}")

    # Load HuggingFace model
    logger.info(f"Going to load the hugging face model {args.model_name}")
    hf_model = AutoModelForCausalLM.from_pretrained(args.model_name, 
                                                    low_cpu_mem_usage=True)
    logger.info(f"Successfully loaded the model {args.model_name}")


    # Save the model
    logger.info('Going to split and save the model')
    save_pretrained_split(hf_model, args.save_path)
    logger.info('Model splitted and saved locally')

    # Load and save tokenizer for the model
    tokenizer = AutoTokenizer.from_pretrained(args.model_name)
    tokenizer.save_pretrained(args.save_path)
    logger.info('Tokenizer saved locally')



"""
Deploys a model on AWS silicon

1. Configuration is read from the configured serving.properties file.
2. A hf_token.txt file is required to download the model from Hugging Face.
"""

# Import necessary libraries
import os
import sys
import stat
import logging
import subprocess
from typing import Dict
from pathlib import Path
from fmbench.scripts import constants

# Set up a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Initialize the platform where this script deploys the model
PLATFORM: str = constants.PLATFORM_SAGEMAKER

# Global constant for the Hugging Face token file
SCRIPT_DIRECTORY: str = os.path.dirname(os.path.realpath(__file__))
HF_TOKEN_FNAME: str = os.path.join(SCRIPT_DIRECTORY, "hf_token.txt")
neuron_script_dir: str = os.path.join(SCRIPT_DIRECTORY, "compile-llm-for-aws-silicon")
shell_script_path: str = os.path.join(neuron_script_dir, "scripts/download_compile_deploy.sh")


def make_executable(file_path: str) -> None:
    """Make the script executable if it is not already."""
    st = os.stat(file_path)
    os.chmod(file_path, st.st_mode | stat.S_IEXEC)
    logger.info("Made script executable: %s", file_path)

def deploy(experiment_config: Dict, role_arn: str) -> Dict:
    """
    Function to deploy the model and create the endpoint
    """
    logger.info("Inside neuron deploy function")

    # Ensure the script is executable
    make_executable(shell_script_path)
    
    requirements_file_path = os.path.join(neuron_script_dir, "requirements.txt")
    command = [sys.executable, '-m', 'pip', 'install', '-r', requirements_file_path]

    try:
        subprocess.check_call(command)
        print("Requirements installed successfully.")
    except subprocess.CalledProcessError as e:
        print(f"Error installing requirements: {e}")

    try:
        model_id = experiment_config['model_id']
        region = experiment_config['region']
        ml_instance_type = experiment_config['instance_type']
        batch_size = experiment_config['ec2']['batch_size']
        image_uri = experiment_config['image_uri']
        num_neuron_cores = experiment_config['ec2']['num_neuron_cores']
        neuron_version = experiment_config['ec2']['neuron_version']
        model_loading_timeout = experiment_config['ec2']['model_loading_timeout']
        prefix = experiment_config['prefix']
        serving_properties = experiment_config['serving.properties']
        s3_bucket = experiment_config['bucket']
        role = experiment_config['sagemaker_execution_role']
        instance_count = experiment_config['ec2']['instance_count']

        logger.info("Model ID: %s", model_id)
        logger.info("Region: %s", region)
        logger.info("Instance Type: %s", ml_instance_type)
        logger.info("Batch Size: %s", batch_size)
        logger.info("Number of Neuron Cores: %s", num_neuron_cores)
        logger.info("Neuron Version: %s", neuron_version)
        logger.info("S3 Bucket: %s", s3_bucket)
        logger.info("Role ARN: %s", role)
        logger.info("Script Path: %s", neuron_script_dir)
        logger.info("Model Loading Timeout: %s", model_loading_timeout)
        logger.info("Initial Instance Count: %s", instance_count)
        hf_token_file_path = Path(HF_TOKEN_FNAME)
        if hf_token_file_path.is_file() is True:
            logger.info(f"hf_token file path: {hf_token_file_path} is a file")
            HF_TOKEN = Path(HF_TOKEN_FNAME).read_text().strip()
        else:
            logger.info(f"hf_token file path: {hf_token_file_path} is not a file")
        logger.info("HF Token is: %s", HF_TOKEN)

    except KeyError as e:
        logger.error("Missing key in experiment_config: %s", e)
        raise
    except FileNotFoundError as e:
        logger.error("File not found: %s", e)
        raise
    except Exception as e:
        logger.error("Error reading configuration: %s", e)
        raise

    logger.info(f"Image uri that is being used: {image_uri}")
    command = [
        shell_script_path,
        HF_TOKEN,
        model_id,
        neuron_version,
        "model_store",
        s3_bucket,
        prefix,
        region,
        role,
        batch_size,
        num_neuron_cores,
        ml_instance_type,
        model_loading_timeout,
        serving_properties,
        neuron_script_dir,
        instance_count, 
        image_uri
    ]
    
    logger.info("Constructed command: %s", command)

    deployment_result = {
        'endpoint_name': None,
        'experiment_name': None,
        'instance_type': None,
        'instance_count': None,
        'deployed': False
    }

    try:
        with open('scripts.log', 'a') as log_file:
            process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, bufsize=1, universal_newlines=True)
    
            for line in process.stdout:
                log_file.write(f"Standard Output: {line}")
                log_file.flush()
                logger.info("Script output: %s", line.strip())

            for line in process.stderr:
                log_file.write(f"Standard Error: {line}")
                log_file.flush()
                logger.error("Script error: %s", line.strip())
            return_code = process.wait()
            
            if return_code == 0:
                logger.info("Script executed successfully")
            else:
                logger.error("Script failed with return code: %d", return_code)
            
            logger.info("Going to read endpoint from endpoint.txt")
            try:
                # Read the endpoint name from the endpoint.txt file
                endpoint_file_path = os.path.join(neuron_script_dir, "endpoint.txt")
                ep_name = Path(endpoint_file_path).read_text().strip()
                logger.info("Endpoint is: %s", ep_name)
                if ep_name:
                    logger.info("Model endpoint running with name: %s", ep_name)
                    deployment_result.update({
                        'endpoint_name': ep_name,
                        'experiment_name': experiment_config['name'],
                        'instance_type': experiment_config['instance_type'],
                        'instance_count': experiment_config['instance_count'],
                        'deployed': True
                    })
                    logger.info("Deployment result: %s", deployment_result)
                else:
                    logger.error("Model endpoint not running!")
            except FileNotFoundError:
                logger.error("File endpoint.txt not found!")
            except Exception as e:
                logger.error("Error reading endpoint.txt: %s", str(e))
                raise

    except subprocess.CalledProcessError as e:
        with open('scripts.log', 'a') as log_file:
            log_file.write("Standard Error:\n")
            log_file.write(e.stderr)
        
        logger.error("Error executing script: %s", e.stderr)
        raise
    except Exception as e:
        logger.error("Unexpected error: %s", str(e))
        raise
    logger.info("Deployment result: %s", deployment_result)
    return deployment_result



"""
vllm specific code
"""
import logging
from fmbench.scripts.inference_containers.utils import (STOP_AND_RM_CONTAINER,
                                                        FMBENCH_MODEL_CONTAINER_NAME)

logging.basicConfig(format='[%(asctime)s] p%(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

def create_script(region, image_uri, model_id, model_name, env_str, privileged_str, hf_token, directory):
    """
    Script for running the docker container for the inference server
    """
    script = f"""#!/bin/sh

        {STOP_AND_RM_CONTAINER}

        # Run the new Docker container with specified settings
        docker run -d {privileged_str} --rm --name={FMBENCH_MODEL_CONTAINER_NAME} --env "HF_TOKEN={hf_token}" --ipc=host -p 8000:8000 {env_str} {image_uri} --model {model_id}

        echo "started docker run in daemon mode"
    """
    return script



import os
import sys
import time
import torch
import logging
import argparse
import torch_neuronx
from transformers import AutoTokenizer
from transformers_neuronx.config import GenerationConfig
from transformers_neuronx.llama.model import LlamaForSampling
from transformers_neuronx import LlamaForSampling, NeuronConfig, GQA, QuantizationConfig

root = logging.getLogger()
if root.handlers:
    for handler in root.handlers:
        root.removeHandler(handler)
logging.basicConfig(format='[%(asctime)s] p%(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

if __name__ == "__main__":
    
    # Define and parse command-line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--action",
        type=str,
        help="What do you want the script do: \"compile\" or \"infer\"?"
    )
    parser.add_argument(
        "--batch-size",
        type=int,
        default=4,
        help="Batch size with which to compile the model, default=4",
    )

    parser.add_argument(
        "--num-neuron-cores",
        type=int,
        default=8,
        help="Number of Neuron cores in the instance in which the model would be run, default=8",
    )

    parser.add_argument(
        "--model-dir",
        type=str,
        help="Directory from where to read the model binaries",
    )
    args = parser.parse_args()
    logger.info(f"args={args}")

    # we will pin cores to 8 for inf2.24xlarge 
    os.environ['NEURON_RT_NUM_CORES'] = str(args.num_neuron_cores)
    os.environ["NEURON_CC_FLAGS"] = "-O3"  ## for best perf
    BATCH_SIZE = args.batch_size
    CONTEXT_LENGTH = 44 # hard coded for sample prompt
    model_dir = args.model_dir
    model_compiled_dir = os.path.join(os.path.dirname(os.path.dirname(model_dir)), "neuronx_artifacts")
    neuron_config = NeuronConfig(on_device_embedding=False,
                                attention_layout='BSH',
                                fuse_qkv=True,
                                group_query_attention=GQA.REPLICATED_HEADS,
                                quant=QuantizationConfig(),
                                on_device_generation=GenerationConfig(do_sample=True))

    if args.action == "compile":
        start = time.perf_counter()
        model = LlamaForSampling.from_pretrained(
                model_dir,
                batch_size=args.batch_size,
                tp_degree=args.num_neuron_cores,
                amp='f16',
                neuron_config=neuron_config,
                n_positions=4096,
                )
        model.to_neuron()
        # save model to the disk
        model.save(model_compiled_dir)
        elapsed = time.perf_counter() - start
        logger.info(f'\nCompilation and loading took {elapsed:.2f} seconds\n')


import json
import time
import logging
import subprocess
import pandas as pd
from threading import Thread

# Setup logging
logging.basicConfig(
    format="[%(asctime)s] %(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s",
    level=logging.INFO,
)
logger = logging.getLogger(__name__)

# Global variables to manage state
data_collection = []
stop_collecting = False
collection_thread = None


def _collect_data(shell_command):
    global data_collection, stop_collecting
    logger.info(f"Starting data collection using command: {shell_command}")

    with subprocess.Popen(
        shell_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as proc:
        try:
            while not stop_collecting:
                output = proc.stdout.readline()
                if output == b"" and proc.poll() is not None:
                    break
                if output:
                    json_data = json.loads(output.decode("utf-8"))

                    # Extract system data
                    system_data = json_data.get("system_data", {})
                    vcpu_usage = system_data.get("vcpu_usage", {}).get(
                        "average_usage", {}
                    )
                    memory_info = system_data.get("memory_info", {})

                    memory_total_gb = memory_info.get("memory_total_bytes", 0) / (
                        1024**3
                    )
                    memory_used_gb = memory_info.get("memory_used_bytes", 0) / (1024**3)
                    swap_total_gb = memory_info.get("swap_total_bytes", 0) / (1024**3)
                    swap_used_gb = memory_info.get("swap_used_bytes", 0) / (1024**3)

                    # Extract neuron_runtime_data
                    neuron_runtime_data = json_data.get("neuron_runtime_data", [])
                    for runtime_entry in neuron_runtime_data:
                        report = runtime_entry.get("report", {})
                        neuroncore_counters = report.get("neuroncore_counters", {})
                        neuron_cores_in_use = neuroncore_counters.get(
                            "neuroncores_in_use", {}
                        )

                        neuron_utilization = {}
                        total_utilization = 0
                        neuroncore_count = 0

                        # Calculate the total utilization and count the number of NeuronCores
                        for core_index, core_data in neuron_cores_in_use.items():
                            utilization = core_data.get("neuroncore_utilization", 0)
                            neuron_utilization[
                                f"neuroncore_{core_index}_utilization"
                            ] = utilization
                            neuron_utilization[f"neuroncore_{core_index}_flops"] = (
                                core_data.get("flops", None)
                            )

                            if utilization is not None:
                                total_utilization += utilization
                                neuroncore_count += 1

                        # Calculate mean utilization as a percentage of total possible utilization
                        if neuroncore_count > 0:
                            mean_utilization = total_utilization / neuroncore_count
                        else:
                            mean_utilization = 0

                        flattened_data = {
                            "vcpu_user": vcpu_usage.get("user"),
                            "vcpu_system": vcpu_usage.get("system"),
                            "vcpu_idle": vcpu_usage.get("idle"),
                            "memory_total_gb": memory_total_gb,
                            "memory_used_gb": memory_used_gb,
                            "swap_total_gb": swap_total_gb,
                            "swap_used_gb": swap_used_gb,
                            "neuroncore_count": neuroncore_count,
                            "total_neuroncore_utilization": total_utilization,
                            "mean_neuroncore_utilization_percentage": mean_utilization,
                            "mean_neuroncore_utilization_total_possible": mean_utilization
                            * neuroncore_count,  # out of 100% * number of cores
                        }

                        # Merge neuron utilization data into the flattened_data
                        flattened_data.update(neuron_utilization)

                        # Append the flattened data to the collection
                        data_collection.append(flattened_data)
        except Exception as e:
            logger.error(f"Error occurred during data collection: {e}")
        finally:
            proc.terminate()
            logger.info("Data collection process terminated.")


def start_collection(shell_command):
    """Starts the data collection in a separate thread."""
    global stop_collecting, collection_thread
    stop_collecting = False
    collection_thread = Thread(target=_collect_data, args=(shell_command,))
    collection_thread.start()
    logger.info("Data collection started in the background.")


def stop_collection():
    """Stops the data collection."""
    global stop_collecting, collection_thread
    stop_collecting = True
    if collection_thread:
        collection_thread.join()
    logger.info("Data collection stopped.")


def get_collected_data():
    """Returns the collected data as a Pandas DataFrame."""
    global data_collection
    logger.info(f"Returning collected data with {len(data_collection)} records.")
    return pd.DataFrame(data_collection)


def reset_collection():
    """Resets the collected data and state."""
    global data_collection, stop_collecting, collection_thread
    data_collection = []
    stop_collecting = False
    collection_thread = None
    logger.info("Data collection has been reset.")



import json
import yaml
import boto3
import logging
from pathlib import Path
from fmbench.utils import load_config
from typing import Dict, Union, List

logger = logging.getLogger(__name__)

# YAML map dictionary for region codes to human-readable location names
REGION_MAP = {
    "us-east-1": "US East (N. Virginia)",
    "us-west-1": "US West (N. California)",
    "us-west-2": "US West (Oregon)",
    "eu-west-1": "EU (Ireland)",
    "ap-southeast-1": "Asia Pacific (Singapore)",
    "ap-southeast-2": "Asia Pacific (Sydney)",
    "ap-northeast-1": "Asia Pacific (Tokyo)",
    "ap-northeast-2": "Asia Pacific (Seoul)",
    "sa-east-1": "South America (Sao Paulo)",
}


def get_ec2_pricing(
    instance_type: str,
    region_code: str,
    operating_system: str = "Linux",
    tenancy: str = "Shared",
) -> float:
    """
    Retrieve on-demand pricing for a specified EC2 instance type and region.

    Parameters:
    - instance_type: e.g., "t3.micro"
    - region_code: e.g., "us-east-1", "us-west-2"
    - operating_system: e.g., "Linux" (default)
    - tenancy: e.g., "Shared" (default) or "Dedicated"

    Returns:
    - Price in USD (float) for the specified instance type and region.
    """
    # Convert region_code to human-readable region name
    region_name = REGION_MAP.get(region_code)
    if not region_name:
        raise ValueError(f"Unsupported region code: {region_code}")

    # Filters for the product
    filters = [
        {"Type": "TERM_MATCH", "Field": "instanceType", "Value": instance_type},
        {"Type": "TERM_MATCH", "Field": "location", "Value": region_name},
        {"Type": "TERM_MATCH", "Field": "operatingSystem", "Value": operating_system},
        {"Type": "TERM_MATCH", "Field": "tenancy", "Value": tenancy},
        {"Type": "TERM_MATCH", "Field": "preInstalledSw", "Value": "NA"},
        {"Type": "TERM_MATCH", "Field": "capacitystatus", "Value": "Used"},
    ]

    # Create a Boto3 client for the Pricing service
    client = boto3.client("pricing", region_name="us-east-1")

    try:
        # Call AWS Pricing API
        response = client.get_products(
            ServiceCode="AmazonEC2", Filters=filters, FormatVersion="aws_v1"
        )

        # Parse response to extract pricing
        for price_item in response["PriceList"]:
            price_data = json.loads(price_item)
            for term in price_data["terms"]["OnDemand"].values():
                for price_dimension in term["priceDimensions"].values():
                    price_per_hour = float(price_dimension["pricePerUnit"]["USD"])
                    return price_per_hour

        # Raise an error if no pricing data is found
        raise ValueError(f"No pricing data found for {instance_type} in {region_code}")

    except boto3.exceptions.Boto3Error as e:
        logger.error(
            f"Boto3 client error while fetching pricing for {instance_type}: {e}"
        )
        raise
    except Exception as e:
        logger.error(
            f"Unexpected error while fetching pricing for {instance_type}: {e}"
        )
        raise


def load_and_update_pricing(
    PRICING_YAML_PATH: Union[Path, str],
    PRICING_FALLBACK_YAML_PATH: Union[Path, str],
    instances: List,
    region_code: str = 'us-east-1',
) -> dict:
    """
    EC2 pricing function which checks if the pricing for a given instance type exists in pricing.yml.
    If it does, it skips fetching. If not, it fetches the pricing using `get_ec2_pricing` and updates pricing.yml.
    If fetching fails, it falls back to the fallback pricing file.

    Args:
        PRICING_YAML_PATH (Union[Path, str]): Path to the pricing YAML file or S3 URI.
        PRICING_FALLBACK_YAML_PATH (Union[Path, str]): Path to the fallback YAML file or S3 URI.
        instances (List): List of instances in experiments to fetch pricing for (e.g., "t2.micro").
        region_code (str): The AWS region code (default is us-east-1).
        
    Returns:
        dict: A dictionary containing updated EC2 pricing data.
    """
    try:
        # Load existing pricing data using the `load_config` function
        pricing_data = load_config(PRICING_YAML_PATH)
        logger.info(f"Loaded pricing data from {PRICING_YAML_PATH}")
    except Exception as e:
        logger.error(f"Error loading pricing YAML from {PRICING_YAML_PATH}: {e}")
        logger.warning("Attempting to load fallback pricing YAML.")
        try:
            pricing_data = load_config(PRICING_FALLBACK_YAML_PATH)
            logger.warning(f"Falling back to {PRICING_FALLBACK_YAML_PATH}")
        except Exception as fallback_error:
            logger.critical(f"Failed to load fallback pricing YAML: {fallback_error}")
            raise FileNotFoundError(
                "Pricing data could not be loaded from any source."
            ) from fallback_error

    for instance_type in instances:
        # Check if the instance type already exists under 'pricing > instance_based'
        if "pricing" in pricing_data and "instance_based" in pricing_data["pricing"]:
            if instance_type in pricing_data["pricing"]["instance_based"]:
                logger.info(f"Pricing for {instance_type} already exists. Skipping fetch.")
                continue
        if "token_based" in pricing_data["pricing"] and instance_type in pricing_data["pricing"]["token_based"]:
            logger.info(f"Token-based pricing for {instance_type} already exists. Skipping fetch.")
            continue  # Skip fetching for this instance type

        # Fetch pricing using `get_ec2_pricing`
        logger.info(
            f"Fetching pricing for instance type: {instance_type} in region: {region_code}"
        )
        
        try:
            price = get_ec2_pricing(instance_type, region_code)
            if price:
                pricing_data["pricing"]["instance_based"][instance_type] = price
                logger.info(f"Fetched pricing for {instance_type}: {price} USD")
            else:
                raise ValueError(
                    f"No pricing data found for {instance_type} in {region_code}"
                )
        except Exception as e:
            logger.error(f"Error fetching pricing: {e}")
            logger.warning(
                "Fetching pricing failed. Falling back to fallback pricing data."
            )
            try:
                pricing_data = load_config(PRICING_FALLBACK_YAML_PATH)
                logger.warning(
                    f"Fallback pricing data loaded from {PRICING_FALLBACK_YAML_PATH}"
                )
            except Exception as fallback_error:
                logger.critical(
                    f"Failed to load fallback pricing YAML after fetch error: {fallback_error}"
                )
                raise FileNotFoundError(
                    "Failed to retrieve pricing and fallback pricing data."
                ) from fallback_error

    return pricing_data
